import sqlite3
import json
from bintang.log import log
from bintang.cell import Cell
from bintang.row import Row
from bintang.table_base import Base_Table


class Sqlite_Table(Base_Table):
    def __init__(self, name, conn, bing):
        super().__init__(name, bing=bing)
        self.conn = conn  # for backend use
        self.conn.row_factory = sqlite3.Row
        self._create_sql_table()


    def __len__(self):
        """ return the length of rows"""
        cursor = self.conn.cursor()
        res = cursor.execute(f'SELECT COUNT(*) FROM {self.name};')
        return res.fetchone()[0]      


    def _create_sql_table(self):
        cur = self.conn.cursor()
        cur.execute("SELECT EXISTS (SELECT 1 FROM sqlite_master WHERE type='table' AND name=:tablename)", {"tablename":self.name})
        ret = cur.fetchone()[0]
        if  ret == 0:
            # cur.execute(f"CREATE TABLE '{self.name}' (id INTEGER PRIMARY KEY NOT NULL, cells JSON)")
            # cur.execute(f"CREATE TABLE '{self.name}__columns__' (id INTEGER PRIMARY KEY NOT NULL, name TEXT COLLATE NOCASE, ordinal_position INTEGER, data_type TEXT, column_size INTEGER, decimal_digits INTEGER, data_props JSON)")
            cur.execute(f"CREATE TABLE '{self.name}' (cells JSON)")
            cur.execute(f"CREATE TABLE '{self.name}__columns__' (columnid INTEGER PRIMARY KEY NOT NULL, name TEXT COLLATE NOCASE, ordinal_position INTEGER, data_type TEXT, column_size INTEGER, decimal_digits INTEGER, data_props JSON)")
        cur.close()

    
    def get_columnid(self, column):
        sql = f"SELECT columnid FROM {self.name}__columns__ where name = ?"
        cursor = self.conn.cursor()
        res = cursor.execute(sql,(column,))
        ret = res.fetchone()
        if ret:
            return ret['columnid']
        else:
            return None
        

    def get_columnids(self, columns):
        sql = f"SELECT columnid FROM {self.name}__columns__ where name IN ({','.join(['?']*len(columns))})"
        cursor = self.conn.cursor()
        res = cursor.execute(sql, columns)
        ret = res.fetchall()
        if ret:
            return [r['columnid'] for r in ret]
        else:
            return None

    
    def _get_last_assigned_ord_pos(self) -> int:
        cursor = self.conn.cursor()
        sql = f"SELECT MAX(ordinal_position) AS MAXOF_ORD_POS FROM {self.name}__columns__;"
        res = cursor.execute(sql)
        ret = res.fetchone()
        if ret:
            if ret['MAXOF_ORD_POS'] is None: # first function call
                return -1
            else:
                return ret['MAXOF_ORD_POS']
        else:
            return -1
    
    
    def _add_column_sql_(self, name, data_type=None, column_size=None):
        # check if the passed name already exists
        columnid = self.get_columnid(name)
        ord_pos = self._get_last_assigned_ord_pos() + 1
        if columnid is None:
            sql = f"INSERT INTO {self.name}__columns__ (name, ordinal_position, data_type, column_size) VALUES (?,?,?,?)"
            params = [name, ord_pos, data_type, column_size]
            cursor = self.conn.cursor()
            cursor.execute(sql, params)
            self.conn.commit()
            cursor.close()
        else:
            log.debug(f'Warning! trying to add existing column "{name}".')

    
    def _make_cell_sql(self,column,value,new_column=True):
        columnid = self.get_columnid(column)
        if columnid is None: # if columnid is None then assume user wants a new column
            if new_column == True:
                self._add_column_sql_(column)
                columnid = self.get_columnid(column) # reassign the columnid
        if columnid is None:
            raise ValueError("Cannot make cell due to None column name.")    
        return Cell(columnid,value) 

    
    def _add_row_sql(self, row, index=None):
        cur = self.conn.cursor()
        # sql = f'INSERT INTO "{self.name}" (id, cells) VALUES (?,?)'
        # cur.execute(sql, (row.id, row.cells)) 
        sql = f'INSERT INTO "{self.name}" (cells) VALUES (?)'
        cur.execute(sql, (row.cells,)) 


    def get_columns(self):
        cursor = self.conn.cursor()
        sql = f"SELECT name FROM {self.name}__columns__ order by ordinal_position;"
        sorted_columns = []
        for row in cursor.execute(sql):
            sorted_columns.append(row['name'])
        return sorted_columns


    def _order_columns_sql(self, columns):
        # determine all columns
        columns_all = [x for x in columns]
        for col in self._get_columns_sql():
            if col not in columns_all:
                columns_all.append(col)
        for i, col in enumerate(columns_all, 10):
            self._update_column_sql(col, ordinal_position=i)


    def _update_column_sql(self,name, data_type=None, column_size=None, ordinal_position=None):
        # check if the passed name already exists
        columnid = self.get_columnid(name)
        if columnid is not None:
            cursor = self.conn.cursor()
            if data_type is not None:
                sql = f"UPDATE {self.name}__columns__ SET data_type=? WHERE id=?;"
                params = [data_type, columnid]
                cursor.execute(sql,params)
            if column_size is not None:
                sql = 'UPDATE {self.name}__columns__ SET column_size=? WHERE id=?;'
                params = [column_size, columnid]
                cursor.execute(sql,params)
            if ordinal_position is not None:
                sql = 'UPDATE {self.name}__columns__ SET ordinal_position=? WHERE id=?;'
                params = [ordinal_position, columnid]
                cursor.execute(sql,params)           






    def insert(self, dict_or_columns, values=None, index=None):
        """ restrict arguments for dict_or_columns insertion for this function as the followings:
        1. a dictionary pass to dict_or_columns param
        2. list values pass to dict_or_columns param and list columns pass to column param.
        """
        if isinstance(dict_or_columns, dict):
            #row = self.make_row()
            row = self.make_row() #Row(0)# dummy idx
            for idx, (col, val) in enumerate(dict_or_columns.items()):
                cell = self._make_cell_sql(col, val)
                row.add_cell(cell) # add to rows
            row.cells = json.dumps({v.columnid: v.value for v in row.cells.values()})
            self._add_row_sql(row, index)
            
                                                  
        elif isinstance(dict_or_columns,list) or isinstance(dict_or_columns,tuple) or isinstance(dict_or_columns,list) or isinstance(dict_or_columns,tuple):
            #row = self.make_row()
            row = self.make_row() #Row(0) # dummy idx
            for idx, col in enumerate(dict_or_columns):
                cell = self._make_cell_sql(col,values[idx])
                row.add_cell(cell) # add to rows
            row.cells = json.dumps({v.columnid: v.value for v in row.cells.values()})
            self._add_row_sql(row, index)
        else:
            raise ValueError("Arg for dict_or_columns set for dictionary or list/tuple of values with list/tuple of columns.")

    
    def _get_columns_withid_sql(self) -> dict:
        # get columnnames from db and return as dict of columnid : columnname
        col_dict = {}
        cur = self.conn.cursor()
        sql = f"SELECT columnid, name FROM {self.name}__columns__ order by ordinal_position;"
        for row in cur.execute(sql):
            col_dict[row['name']] = row['columnid'] 
        return col_dict


    def _gen_cells_dict(self, cells: str) -> dict:
        return {int(k):v for k,v in json.loads(cells).items()}    
    
    
    def _gen_row_asdict_sql(self, cells: str, columns: list) -> dict:
        db_cols_withid = self._get_columns_withid_sql()
        user_cols = {k:v for k,v in db_cols_withid.items() if k in columns}
        cells_dict = self._gen_cells_dict(cells)
        row_dict = {}
        for col in columns:
            if user_cols[col] in cells_dict:
                row_dict[col] = cells_dict[user_cols[col]] 
            else:
                row_dict[col] = None
        return row_dict


    def _iter_row_asdict_sql(self,columns):
        # get columnames
        db_cols_withid = self._get_columns_withid_sql()
        user_cols = {k:v for k,v in db_cols_withid.items() if k in columns} #refine columns
       
        cur = self.conn.cursor()
        #sql = "SELECT rowid, cells FROM {}".format(self.name) # expose rowid
        sql = "SELECT cells FROM {}".format(self.name)
        for idx, row in enumerate(cur.execute(sql),start=1):
            # debug cells_dict = json.loads(row["cells"])
            # log.debug(cells_dict)
            cells_dict = self._gen_cells_dict(row['cells'])
            row_asdict = {}
            for col in columns:
                if user_cols[col] in cells_dict:
                    row_asdict[col] = cells_dict[user_cols[col]]
                else:
                    row_asdict[col] = None
            # yield row['rowid'], row_asdict
            yield idx, row_asdict


    def _iter_row_aslist_sql(self,columns):
        # get columnames
        db_cols_withid = self._get_columns_withid_sql()
        user_cols = {k:v for k,v in db_cols_withid.items() if k in columns} #refine columns
       
        cur = self.conn.cursor()
        # sql = "SELECT rowid, cells FROM {}".format(self.name)
        sql = "SELECT cells FROM {}".format(self.name)
        for idx, row in enumerate(cur.execute(sql),start=1):
            # debug cells_dict = json.loads(row["cells"])
            # log.debug(cells_dict)
            cells_dict = self._gen_cells_dict(row['cells'])
            row_aslist = []
            for col in columns:
                if user_cols[col] in cells_dict:
                    row_aslist.append(cells_dict[user_cols[col]])
                else:
                    row_aslist.append(None) #???? this is strange line of code. revisit!!
            # yield row["idx"], row_aslist    
            yield idx, row_aslist
    
    def iterrows(self, 
                 columns: list=None, 
                 row_type: str='dict', 
                 where=None, 
                 rowid: bool=False):

        # validate user's args
        if columns is not None:
            ## need to work on validation from diff types of table
            db_columns = self._get_columns_sql()
            missing_cols = []
            for col in columns:
                if col not in db_columns:
                    missing_cols.append(col)
            if len(missing_cols) > 0:
                raise ValueError ('Error! Column {} not found.'.format(','.join(missing_cols)))
                

        if columns is None:
            columns = self.get_columns() # assign all available column names
        else:
            columns = self.validate_columns(columns) # assign all available column names after validation
            
        if row_type == 'list':
            for idx, row in self._iter_row_aslist_sql(columns):
                yield idx, row 
                    
        else: # assume row_type is dict 
            for idx, row in self._iter_row_asdict_sql(columns):
                yield idx, row


    def get_row_asdict(self, idx, columns=None):
        if columns is None:
            columns = self.get_columns() # assign all available column names
        else:
            columns = self.validate_columns(columns) # assign all available column names after validation
        cursor = self.conn.cursor()
        sql = "SELECT cells FROM {} WHERE idx=?".format(self.name)
        res = cursor.execute(sql, (idx,))
        ret = res.fetchone()
        if ret:
            print(ret['cells'])
            return self._gen_row_asdict_sql(ret['cells'], columns)
        else:
            return None


    def get_row_aslist(self, idx, columns=None):
        if columns is None:
            columns = self.get_columns() # assign all available column names
        else:
            columns = self.validate_columns(columns) # assign all available column names after validation
        cursor = self.conn.cursor()
        sql = "SELECT cells FROM {} WHERE idx=?".format(self.name)
        res = cursor.execute(sql, (idx,))
        ret = res.fetchone()
        if ret:
            return list(self._gen_row_asdict_sql(ret['cells'], columns).values())
        else:
            return None                

    
    def update_row(self, idx, column, value):
        """ using json_set so only update for a specific key"""
        cursor = self.conn.cursor() 
        columnid = self.get_columnid(column)
        sql_update = f"UPDATE {self.name} SET cells = (SELECT json_set({self.name}.cells, '$.{columnid}', ?)) where idx=?;"
        params = [value, idx]
        cursor.execute(sql_update, params)


    def _update_row_sql_bycolid(self, idx, columnid, value):
        """ using json_set so only update for a specific key and accept columnid
        Note: no performance increase even not calling get_columnid_sql()"""
        cursor = self.conn.cursor() 
        sql_update = f"UPDATE {self.name} SET cells = (SELECT json_set({self.name}.cells, '$.{columnid}', ?)) where idx=?;"
        params = [value, idx]
        cursor.execute(sql_update, params)    
                            




class Cell(object):
    """Define a cell object
    """
    __slot__ = ('columnid', 'value')
    def __init__(self,columnid,value = None):
        self.columnid = columnid
        self.value = value

    def __repr__(self):
        return "{}(columnid:{}, value: {})".format(__class__.__name__, str(self.columnid), str(self.value))
    

class Cell_Path_List(Cell):
    def __init__(self,columnid, path_list, value):
        super().__init__(columnid, value)
        self.path_list = path_list
        self.is_key = False

    def __repr__(self):
        return "{}(columnid:{}, value:{}, path_list:{}, get_column():{}, is_parent_key:{})".format\
            (__class__.__name__, str(self.columnid), str(self.value)\
            ,str(self.path_list), self.get_column(), self.is_key)    
        
    def gen_path_list_norowid(self):
        pathl_norowid = [x for x in self.path_list if not isinstance(x, int)]
        return pathl_norowid
     
    def get_column(self):
        pathl_norowid = self.gen_path_list_norowid()
        colname = pathl_norowid[-1]
        if pathl_norowid.count(colname) > 1: # resolve name if table and column would use the same name
            return colname + '_' + colname
        else:
            return pathl_norowid[-1]


import unicodedata
class Column:
    """class to define column object 
    """
    __slots__ = ('id','name','data_type','required','min_value','max_value','min_length','max_length','ordinal_position','column_size','decimal_digits','data_props')
    def __init__(self,name):
        self.id = None
        self.name = name
        self.data_type = None  # for validation
        self.required = False # for validation
        self.min_value = None # for validation for int and float
        self.max_value = None # for validation for int and float
        self.min_length = None # for validation for str
        self.max_length = None # for validation for str
        self.ordinal_position = 0
        self.column_size = 0 # the max. get this on request
        self.decimal_digits = 0 # get this on request
        self.data_props = {}
        #self.data_types = [] # get this on request. its common data type more than one (inconsistent data)
        
        
    
    def __repr__(self):
        return f'id:{self.id}, name: {self.name}, data_type: {self.data_type}, required: {self.required}, min_value: {self.min_value}, max_value: {self.max_value}, min_length: {self.min_length}, max_length: {self.max_length}'

    def get_name_uppercased(self):
        return self.name.upper()
    
    # def get_name_normalize_caseless(self):
    #     return unicodedata.normalize("NFKD", self.name.casefold())
    
    def get_greatest_column_size_data_prop(self):
        """data props can hold diffrent data type
        depending data source provided."""
        column_size = 0
        data_prop = None
        for k, v in self.data_props.items():
            if v['column_size'] > column_size:
                 column_size = v['column_size']
                 data_prop = {k: v}
        return data_prop
    
    def has_string_data_type(self):
        #print('hello')
        for dt in self.data_props.keys():
            #print('  dt',dt)
            if dt == 'str':
                return True
        return False


    def get_data_types(self):
        # data type from populated data
        data_types = []
        for dt in self.data_props.keys():
            data_types.append(dt)
        return data_types


from contextlib import contextmanager

class Row:
    """Define a row as a collection of cells
    """
    __slots__ = ('id', 'cells') 
    def __init__(self,id):
        self.id = id
        self.cells = {}


    def __repr__(self):
        return '{}(id:{}, cells:{})'.format(__class__.__name__, self.id, self.cells)


    def add_cell(self,cell):
        self.cells[cell.columnid] = cell # add a cell into the cells dictionary and use columnid as the key


    def get_cell(self,columnid):
        cell = self.cells[columnid]
        #print("test direct...",cell.get_cell())
        return cell


    def get_value(self,columnid):
        if columnid in self.cells:
            return self.cells[columnid].value


    def get_values(self,columnids):
        row_values = []
        for columnid in columnids:
            cell_value = None
            if columnid in self.cells:
                cell_value = self.cells[columnid].value
            row_values.append(cell_value)
        return row_values


class Row_Table_Path(Row):
    def __init__(self,id):
        super().__init__(id)
        self.tablepath: str | None = None

    # DEPRECATED use super() instead
    # def __init__(self, id):
    #     Row.__init__(self, id)
    #     self.tablepath = None    

    def __repr__(self):
        return '{}(tablepath:{} id:{}, cells:{})'.format(__class__.__name__, self.tablepath, self.id, self.cells)            


from typing import Union
import os
import json
import copy
import unicodedata
from bintang.table_base import Base_Table
from bintang.table_mem import Memory_Table, Path_Table
from bintang.table_sqlite import Sqlite_Table
from bintang.table_fsql import From_SQL_Table
from bintang.table_fcsv import From_CSV_Table
from bintang import iterdict
from pathlib import Path
from bintang.log import log
import unicodedata
import datetime


def default_stringify(value):
    """ convert value to string, normalize it and lower case it.
        This is used to normalize the value before comparing it.
        It aimes to cover default process that SequenceMatcher does not handle.
    """
    if isinstance(value, str):
        value = value.strip()
        value =unicodedata.normalize("NFKD", value.casefold()) #experiment
        return value.lower()
    else:
        return str(value)


# define get_fuzzy_ratio() to use rapidfuzz or difflib depending on package availability
try:
    from rapidfuzz import fuzz , process, utils
    def get_fuzzy_ratio(value1, value2, default_process=True):
        """ get string similarity between two values.
            using rapidfuzz package only
        """
        if value1 is None or value2 is None:
            return 0  
        elif default_process:
            return  round(fuzz.ratio(value1, value2, processor=utils.default_process) /100.0 ,4)
        else:
            return  round(fuzz.ratio(value1, value2) /100.0, 4)
except ImportError as e:
    log.warning(f'warning! bintang.get_fuzzy_ratio() could not find rapidfuzz, difflib.SequenceMatcher would be used.')
    # define get_fuzzy_ratio to use difflib
    from difflib import SequenceMatcher
    def get_fuzzy_ratio(value1, value2, default_process=True):
        """ get string similarity between two values.
            using python difflib's SequenceMatcher only
        """
        if default_process == True:
            value1 = default_stringify(value1)
            value2 = default_stringify(value2)
            return round(SequenceMatcher(None, value1, value2).ratio(), 4)
        else:
            return round(SequenceMatcher(None, value1, value2).ratio(), 4)


def match_case(value1, value2):
    "Ascii case sensitive matching"
    if value1 == value2:
        return True
    

def match_caseless(value1, value2):
    "Ascii ase insensitive matching"
    if isinstance(value1, str) and isinstance(value2, str):
        if value1.lower() == value2.lower():
            return True
    elif isinstance(value1, str) or isinstance(value2, str):
        if str(value1) == str(value2):
            return True    
    else:
        if value1 == value2:
            return True    



def _normalize_caseless(string):
    return unicodedata.normalize("NFKD", string.casefold())


def match(value1, value2):
    """Unicode case insensitive matching.
    the ultimate goal is to get result regardless data type"""
    if isinstance(value1, str) and isinstance(value2, str):
        if _normalize_caseless(value1) == _normalize_caseless(value2):
            return True
    elif isinstance(value1, str) or isinstance(value2, str): # 1 vs '1' is True
        if str(value1) == str(value2):
            return True
    else:
        if value1 == value2:
            return True  

    
def get_similar_values(value, similar_values, min_ratio=0.6):
        res = []
        for col in similar_values:
            ratio = get_fuzzy_ratio(col, value)
            print(col,value)
            print(ratio)
          
            if ratio >= min_ratio:
                res.append((col,ratio))
        res_sorted = sorted(res, key=lambda tup: tup[1], reverse=True)
        return [x[0] for x in res_sorted] # just extract the name, not ratio


def get_wb_type_toread(wb):
    """
        get workbook type so read_excel() will know which func/attrb to use.
    """
    if str(type(wb)) == "<class 'openpyxl.workbook.workbook.Workbook'>":
        return 'openpyxl'
    elif str(type(wb)) == "<class 'xlrd.book.Book'>":
        return 'xlrd'
    else:
        raise ValueError('Sorry, for reading excel file, only openpyxl or xlrd workbook accepted!')
    

def get_wb_type_towrite(wb):
    """
        get workbook type so to_excel() know which func/attrb to use.
    """
    if str(type(wb)) == "<class 'openpyxl.workbook.workbook.Workbook'>":
        return 'openpyxl'
    elif str(type(wb)) == "<class 'xlwt.Workbook.Workbook'>":
        return 'xlwt'
    else:
        raise ValueError('Sorry, for writing excel file, only openpyxl or xlwt workbook accepted!')


def gen_ts_string(fmt: str = None) -> str:
    """
    Generates a timestamp string. Defaults to YYYYMMDDTHHMMSS 
    (ISO 8601 basic format) if no format is provided.
    """
    now = datetime.datetime.now()
    if fmt is None:
        return now.strftime('%Y%m%dT%H%M%S')
    return now.strftime(fmt)


def gen_ts_path(file_path: str, fmt: str = '%Y%m%dT%H%M%S') -> str:
    """
    Appends a timestamp to a filename while preserving the directory path.
    Example: 'c:/data/source.csv' -> 'c:/data/source_20260314T093500.csv'
    """
    original_path = Path(file_path)
    timestamp = datetime.datetime.now().strftime(fmt)
    
    # original_path.parent gives you "c:\data"
    # original_path.stem gives you "incoming_data"
    # original_path.suffix gives you ".csv"
    new_filename = f"{original_path.stem}_{timestamp}{original_path.suffix}"
    
    # joinpath puts the folder and the new filename back together correctly
    return str(original_path.parent.joinpath(new_filename))


class Bintang():
    def __init__(self, name=None, backend=None):
        self.name = name
        # self.parent = 'dad'
        self.__tables = {} # this must be a dict of id:table object
        self.__last_assigned_tableid= 0 # to keep track of last assigned tableid

    def __getitem__(self, tablename: str) -> Memory_Table: # subscriptable version of self.get_table()
        tableid = self.get_tableid(tablename)
        if tableid is None:
            tablenames = self.get_tables()
            similar_tables = get_similar_values(tablename, tablenames)
            if len(similar_tables) > 0:
                raise ValueError ('could not find table {}. Did you mean {}?'.format(repr(tablename),' or '.join(similar_tables)))
            else:
                raise ValueError ('could not find table {}.'.format(repr(tablename)))
        else:
            return self.__tables[tableid]       

    
    def __repr__(self):
        rb = {}
        rb['name'] = self.name
        table = []
        for tablename in self.get_tables():
            table.append(tablename)
        rb['tables'] = table
        return json.dumps(rb, indent=2)


    def create_table(self, name: str, 
                     columns: list=None) -> Memory_Table:
        """ create a Memory_Table under bintang object
        name: Name of the table
        columns: List of columns (optional)"""
        tobj = Memory_Table(name, bing=self)
        self.add_table(tobj)
        if columns is not None:
            for column in columns: 
                tobj.add_column(column)
        return tobj        

    
    def drop_table(self, name):
        tableid = self.get_tableid(name)
        if tableid is None:
            tables = self.get_tables()
            similar_tables = get_similar_values(name, tables)
            raise ValueError ('could not find table {}. Did you mean {}?'.format(repr(name),' or '.join(similar_tables)))
        else:
            del self.__tables[tableid]


    def create_path_table(self, name, columns=None):
        tobj = Path_Table(name, bing=self)
        self.add_table(tobj)
        if columns is not None:
            for column in columns: # add column
                tobj.add_column(column)            
        
        
    def get_tableid(self, tablename):
        for id, table in self.__tables.items():
            if tablename.upper() == table.name.upper(): # compare uppercased to ensure case insensitivity
                return id
        return None


    def get_tables(self):
        tables = []
        for table in self.__tables.values():
            tablename = table.name
            tables.append(tablename)    
        return tables


    def get_tablepaths(self):
        pathnames = []
        for table in self.__tables.values():
            pathname = table.path
            pathnames.append(pathname)    
        return pathnames    

    
    def get_columns(self, tablename):
        return self.get_table(tablename).get_columns()
    

    def add_table(self, table: Union[Memory_Table, Path_Table, Sqlite_Table, From_SQL_Table, From_CSV_Table]): 
        tableid = self.get_tableid(table.name)
        if tableid is None:
            tableid = self.__last_assigned_tableid + 1
            self.__tables[tableid] = table
            self.__last_assigned_tableid = self.__last_assigned_tableid + 1 # update it
        elif tableid is not None:
            raise ValueError('Table {} already exists.'.format(table.name))


    def rename_table(self, name, new_tablename):
        # check if name already exists
        if name in self.get_tables():
            for id, table in self.__tables.items():
                if name == table.name:
                    self.__tables[id].name = new_tablename
        else:
            raise ValueError('Tablename {} does not exist.'.format(name))

    
    def get_table(self, name):
        tableid = self.get_tableid(name)
        if tableid is not None:
            return self.__tables[tableid]
        else:
            return None


    def copy_table(self,source_tablename, destination_tablename):
        destination_table = copy.deepcopy(self.get_table(source_tablename))
        destination_table.name = destination_tablename
        self.add_table(destination_table)
        
    
    def drop_column(self,tablename,column):
        self.__tables[tablename].drop_column(column)


    def insert(self,tablename,columns,values):
        self.get_table(tablename).insert(columns,values)
        

    def _get_cells_bycolumns(self,tablename,row,columns):
        _cells = {} # to hold the results cells
        for name in columns:
            columnid = self.__tables[tablename].get_columnid(name)
            # only assign if columnid exists to avoid a KeyError
            if columnid in row.cells:
                _cells[columnid] = row.cells[columnid]
        return _cells

    
    def delete_row(self, tablename, index):
        self.get_table(tablename).delete_row(index)

    
    def delete_rows(self, tablename, indexes):
        self.get_table(tablename).delete_rows(indexes)


    def _print(self):
        tobj = Table('Columns Info')
        row_dict = {}
        for tab in self.get_tables():
            self[tab].set_data_props() # it sets 
            row_dict['table'] = tab
            for col in self[tab].get_columns():
                cobj = self[tab].get_column_object(col)
                #row_dict['id'] = cobj.id
                row_dict['column'] = cobj.name
                row_dict['ordinal_position'] = cobj.ordinal_position
                row_dict['data_type'] = cobj.data_type
                row_dict['column_size'] = cobj.column_size
                row_dict['decimal_digits'] = cobj.decimal_digits
                row_dict['data_props'] = cobj.data_props
                tobj.insert(row_dict)
        tobj.print()


    def print(self):
        tobj = Memory_Table('Tables')
        for tab in self.get_tables():
            row_dict = {}
            row_dict['Bintang'] = self.name
            row_dict['Table'] = tab
            tobj.insert(row_dict)
        tobj.print()
 

    def read_excel(self, wb, sheetnames=None):
        wb_type = get_wb_type_toread(wb)
        # validate sheetnames
        # sheetnames_lced = {x.lower(): x  for x in wb.sheetnames}
        # validate sheetname
        if wb_type == 'openpyxl':
            sheetnames_lced = {x.lower(): x  for x in wb.sheetnames}
        else: # assume xlrd
            sheetnames_lced = {x.lower(): x  for x in wb.sheet_names()}
        if sheetnames is not None: # user specify sheets
            for sheetname in sheetnames:
                if sheetname.lower() not in sheetnames_lced:
                    # get similar sheetname
                    if wb_type == 'openpyxl':
                        similar_sheetnames = get_similar_values(sheetname, wb.sheetnames)
                    else: # assume xlrd
                        similar_sheetnames = get_similar_values(sheetname, wb.sheet_names())
                    raise ValueError ('could not find sheetname {}. Did you mean {}?'.format(repr(sheetname),' or '.join(similar_sheetnames)))
                self.create_table(sheetname)
                self[sheetname].read_excel(wb, sheetname)
        else: # exctract all sheets and create all tables
            for ws in wb:
                sheetname = ws.title
                self.create_table(sheetname)
                self[sheetname].read_excel(wb, sheetname)


    def read_dict(self, dict_obj, tablepaths=None, root='/', path_sep='/'):
        debug = False
        for tprow in iterdict.iterdict(dict_obj, tablepaths=tablepaths, root=root, path_sep=path_sep):
            # if debug:
            #     print("\n---------------------in bintang---------------------")
            #     print(tprow)
            
            for k,v in tprow.cells.items():
                # print(k,'->',v)
                
                # create a table if not created yet
                if tprow.tablepath not in self.get_tables():
                    self.create_path_table(tprow.tablepath)
                
                # upsert this row
                self.get_table(tprow.tablepath).upsert_table_path_row(tprow)
            # if debug:
            #     print("---------------------out bintang--------------------")


    def read_json(self, json_str, tablepaths=None):
        dict_obj = json.loads(json_str) # parse json_str
        self.read_dict(dict_obj, tablepaths) 


    def set_child_tables(self):
        for tab1 in self.get_tablepaths():
            print('table name',tab1)
            # print(self[tab1].get_path_aslist())
            tab1_pathlist = self[tab1].get_path_aslist()
            for tab2 in self.get_tablepaths():
                tab2_pathlist = self[tab2].get_path_aslist()
                if len(tab2_pathlist) - len(tab1_pathlist) == 1:
                    # this is a possible child, not grand child
                    matches = 0
                    for i in range(len(tab1_pathlist)):
                        if tab1_pathlist[i] == tab2_pathlist[i]:
                            matches += 1
                    if matches == len(tab1_pathlist):
                        # this is a child
                        self[tab1].children.append(tab2)      


    def read_sql(self, conn, tablename, sql_str, params=None ):
        self.create_table(tablename)
        cursor = conn.cursor()
        if params is not None:
            cursor.execute(sql_str, params)
        else:
            cursor.execute(sql_str)
        columns = [col[0] for col in cursor.description]
        for row in cursor.fetchall():
            self.insert(tablename, columns, row)


    def _add_lcell(self, lidx, ltable, outrow, out_table, out_lcolumns, rowid):
        for k, v in self[ltable].get_row_asdict(lidx,rowid=rowid).items():
            if out_lcolumns is None:
                # incude all columns
                cell = out_table.make_cell(k,v)
                outrow.add_cell(cell)
            if out_lcolumns is not None:
                # include only the passed columns
                if k in out_lcolumns:
                    cell = out_table.make_cell(k,v)
                    outrow.add_cell(cell)
        return outrow            


    def _add_rcell(self, ridx, rtable, outrow, out_table, out_rcolumns, rcol_resolved, rowid):
        for k, v in rtable.get_row_asdict(ridx,rowid=rowid).items():
            if out_rcolumns is None:
                cell = out_table.make_cell(rcol_resolved[k],v)
                outrow.add_cell(cell)
            if out_rcolumns is not None:
                # include only the passed columns
                if k in out_rcolumns:
                    cell = out_table.make_cell(rcol_resolved[k],v)
                    outrow.add_cell(cell)
        return outrow


    def _resolve_join_columns (self,ltable, rtable_obj, rowid=False):
        # resolve any conflict column by prefixing its tablename
        # notes: conflict column occurs when the same column being used in two joining table.
        rcol_resolved = {}
        if rowid:
            rcol_resolved['rowid_'] = ltable + '_' + 'rowid_'
        lcolumns = self[ltable].get_columns()
        for col in rtable_obj.get_columns():
            if col in lcolumns:
                rcol_resolved[col] = rtable_obj.name + '_' + col
            else:
                rcol_resolved[col] = col
        return rcol_resolved  


    def innerjoin(self
                ,ltable: str #, lkeys
                ,rtable: str | Memory_Table #, rkeys
                ,on: list[tuple] # list of lkey & r key tuple
                ,into: str | None =None
                ,out_lcolumns: list | None = None
                ,out_rcolumns: list | None = None
                ,rowid=False) -> Memory_Table:
        
        rtable_obj = None
        if isinstance(rtable, Memory_Table):
            rtable_obj = rtable
            rtable = rtable.name
        else:
            rtable_obj = self[rtable]
        # validate input eg. column etc
        lkeys = [x[0] for x in on] # generate lkeys from on (1st sequence)
        rkeys = [x[1] for x in on] # generate rkeys from on (2nd sequence)
        lkeys = self[ltable].validate_columns(lkeys)
        #rkeys = self[rtable].validate_columns(rkeys)
        rkeys = rtable_obj.validate_columns(rkeys)
        if out_lcolumns is not None:
            out_lcolumns = self[ltable].validate_columns(out_lcolumns)
        if out_rcolumns is not None:
            # out_rcolumns = self[rtable].validate_columns(out_rcolumns)
            out_rcolumns = rtable_obj.validate_columns(out_rcolumns)

        # resolve columns conflicts
        rcol_resolved = self._resolve_join_columns(ltable, rtable_obj, rowid)
        # create an output table
        out_table = into if into is not None else 'innerjoin'
        out_tobj = Memory_Table(out_table)
        # self.create_table(out_table)
        # out_tobj = self.get_table(out_table)

        # for debuging create merged table to store the matching rowids
        #merged = self.create_table("merged",["lrowid","rrowid"])

        numof_keys = len(on) #(lkeys)
        # loop left table
        # for lidx, lrow in self.iterrows(ltable, columns=lkeys, rowid=rowid):
        for lidx, lrow in self[ltable].iterrows(columns=lkeys, rowid=rowid):    
            # loop right table
            #for ridx, rrow in self.iterrows(rtable_obj, columns=rkeys, rowid=rowid):
            for ridx, rrow in self[rtable].iterrows(columns=rkeys, rowid=rowid):
                matches = 0 # store matches for each rrow
                # compare value for any matching keys, if TRUE then increment matches
                for i in range(numof_keys): 
                    if match(lrow[lkeys[i]], rrow[rkeys[i]]):
                        matches += 1 # incremented!
                if matches == numof_keys: # if fully matched, create the row & add into the output table
                    #debug merged.insert(["lrowid","rrowid"], [lrow["_rowid"], rrow["_rowid"]])
                    #and add the row to output table
                    outrow = out_tobj.make_row()
                    # add cells from left table
                    outrow = self._add_lcell(lidx, ltable, outrow, out_tobj, out_lcolumns, rowid)
                    # add cells from right table
                    outrow = self._add_rcell(ridx, rtable_obj, outrow, out_tobj, out_rcolumns, rcol_resolved, rowid)   
                    out_tobj.add_row(outrow)
        #debug merged.print() 
        return out_tobj
                

    def leftjoin(self,ltable: str
                ,rtable: str
                ,lkeys, rkeys
                ,out_lcolumns=None
                ,out_rcolumns=None
                ,rowid=False):

        # validate input eg. column etc
        lkeys = self[ltable].validate_columns(lkeys)
        rkeys = self[rtable].validate_columns(rkeys)
        if out_lcolumns is not None:
            out_lcolumns = self[ltable].validate_columns(out_lcolumns)
        if out_rcolumns is not None:
            out_rcolumns = self[rtable].validate_columns(out_rcolumns)
        

        # create an output table 
        out_table = ltable + rtable
        self.create_table(out_table)
        out_tobj = self.get_table(out_table)

        # for debuging create merged table to store the matching rowids
        #merged = self.create_table("merged",["lrowid","rrowid"])

        numof_keys = len(lkeys)
        for lidx, lrow in self.iterrows(ltable, columns=lkeys, rowid=rowid):
            outrow = out_tobj.make_row()
            # add cells for ltable
            outrow = self._add_lcell(lidx, ltable, outrow, out_table, out_lcolumns, rowid)
            for ridx, rrow in self.iterrows(rtable, columns=rkeys, rowid=rowid):
                matches = 0 # store matches for each rrow
                # evaluate any matching keys, if so increment matches
                for i in range(numof_keys):
                    if match(lrow[lkeys[i]] == rrow[rkeys[i]]):
                        matches += 1 # increment
                if matches == numof_keys: # if fully matched, create the row & add into the output table
                    #debug merged.insert(["lrowid","rrowid"], [lrow["_rowid"], rrow["_rowid"]])
                    # add cells for rtable
                    outrow = self._add_rcell(ridx, rtable, outrow, out_table, out_rcolumns, ltable, rowid)   
            out_tobj.add_row(outrow)             
        #debug merged.print() 
        return out_tobj
    

    def create_sqlite_table(self, name, conn):
        tobj = Sqlite_Table(name, conn, bing=self) # create a tobj object
        self.add_table(tobj)
        return tobj
    
    
    def create_linked_table(self, name, conn, sql_str=None, params=None ):
        tobj = From_SQL_Table(name, bing=self) # create a tobj object
        tobj.type = 'FROMSQL'
        tobj.conn = conn
        if sql_str is None:
            tobj.sql_str = "SELECT * FROM {}".format(name)
        else:
            tobj.sql_str = sql_str
        tobj.params = params    
        self.add_table(tobj)


    def create_sql_linked_table(self, name, conn, sql_str=None, params=None ):
        """ this is an alias to create_linked_table() """
        tobj = From_SQL_Table(name, bing=self) # create a tobj object
        tobj.conn = conn
        if sql_str is None:
            tobj.sql_str = "SELECT * FROM {}".format(name)
        else:
            tobj.sql_str = sql_str
        tobj.params = params    
        self.add_table(tobj)


    def create_csv_linked_table(self, name, filepath, delimiter=',', quotechar='"', header_row=1):
        tobj = From_CSV_Table(name, filepath, delimiter=delimiter, quotechar=quotechar, header_row=header_row, bing=self) # create a tobj object
        tobj.filepath = filepath
        tobj.delimiter = delimiter
        tobj.quotechar = quotechar
        tobj.header_row = header_row
        self.add_table(tobj) 
  


    


import subprocess
import glob
import pytest
import os
import sys


example_scripts = glob.glob(os.path.join("examples", "*.py"))

@pytest.mark.parametrize("script_path", example_scripts)
def test_example_execution(script_path):
    """
    to test each file under example folder
    """
    print(script_path)
    # Get the current system environment variables
    env = os.environ.copy()
    
    # Add the 'src' folder to the PYTHONPATH so the script can find 'bintang'
    # This works regardless of whether you are on Windows or Mac/Linux
    src_path = os.path.abspath("src")
    env["PYTHONPATH"] = src_path + os.pathsep + env.get("PYTHONPATH", "")

    result = subprocess.run(
        [sys.executable, script_path], # sys.executable ensures it uses the same python as pytest
        capture_output=True, 
        text=True,
        env=env  # Pass the updated environment here
    )
    
    assert result.returncode == 0, f"Example {script_path} failed!\nError: {result.stderr}"


from bintang import Bintang
import pytest
import datetime

@pytest.fixture(scope="session")
def bt():
    """ fixture for bintang object
    """
    bt = Bintang("test_bt")
    return bt

def test_bt_init(bt):
    assert bt.name == "test_bt"
    assert isinstance(bt, Bintang)

def test_create_table(bt):
    bt.create_table("Person")
    assert "Person" in bt.get_tables()

def test_insert_person_table(bt):
    """ test inserting rows into Person table
        the insert method should accept both dict and list/tuple
    """

    bt['Person'].insert({'id': 1, 'name': 'John', 'surname': 'Smith', 'address': '1 Station St'})

    expected_columns = ('id', 'name', 'surname', 'address')
    assert bt['Person'].get_columns() == expected_columns

    # insert another record with new column, hobby.
    bt['Person'].insert(['id','name','surname','hobby','address'], [2,'Jane','Brown','Digging','8 Parade Rd'])

    expected_columns = ('id', 'name', 'surname', 'address', 'hobby')
    assert bt['Person'].get_columns() == expected_columns # hobby column should be added

    p = bt['Person']  # set p as the table bt['Person'] alias so the codes more readable
    p.insert(['id','name','surname','Address'], [3,'Okie','Dokie','7 Ocean Rd'])
    p.insert(['id','name','hobby','Address'], [4,'Maria','Digging','7 Heaven Ave'])
    p.insert(['id','name','hobby','Address'], [5,'Bing','Digging',None])

    expected_rows = [
        {'id': 1, 'name': 'John', 'surname': 'Smith', 'address': '1 Station St', 'hobby': None},
        {'id': 2, 'name': 'Jane', 'surname': 'Brown', 'address': '8 Parade Rd', 'hobby': 'Digging'},
        {'id': 3, 'name': 'Okie', 'surname': 'Dokie', 'address': '7 Ocean Rd', 'hobby': None},
        {'id': 4, 'name': 'Maria', 'surname': None, 'address': '7 Heaven Ave', 'hobby': 'Digging'},
        {'id': 5, 'name': 'Bing', 'surname': None, 'address': None, 'hobby': 'Digging'}
    ]

    assert len(p) == len(expected_rows)

    for idx, row in p.iterrows():
        assert row == expected_rows[idx -1] # decreased idx to match expected_rows index list.

def test_get_and_update_table(bt):
    p = bt['Person']
    p.update('age', 10, where=lambda row: row['name']=='Maria')
    assert 'age' in p.get_columns()
    assert p.get_row_asdict(4)['age'] == 10

    p.rename_column('name', 'first_name')
    assert 'name' not in  p.get_columns() # 'name must not be in the columns list
    assert 'first_name' in p.get_columns() # 'first_name must be in the column list

    assert p.get_value('first_name', where=lambda row: row['id']==1) == 'John'



from bintang.table_base import Base_Table
from bintang.log import log
import csv

class From_CSV_Table(Base_Table):
    def __init__(self, name, filepath, bing=None, delimiter=',', quotechar='"', header_row=1):
        super().__init__(name, bing=bing)
        self.filepath = filepath
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.header_row = header_row
        self.columns = self.get_columns()


    def __len__(self):
        """ return the length of rows"""
        with open(self.filepath, newline='') as f:
            reader = csv.reader(f, delimiter=self.delimiter, quotechar=self.quotechar)
            row_count = len(list(reader)) -1
        return row_count


    def get_columnid(self, column):
        pass


    def get_columnids(self, columns):
        pass    


    def get_columns(self):
        with open(self.filepath, newline='') as f:
            reader = csv.reader(f, delimiter=self.delimiter, quotechar=self.quotechar)
            # determine columns
            columns = []
            for rownum, row in enumerate(reader, start=1):
                if rownum == self.header_row:
                    columns = [col for col in row] # add all columns
                    f.seek(0) # return to BOF
            return tuple(columns)
                    

    

    
    def iterrows(self, 
                 columns: list | tuple | None = None, 
                 row_type: str='dict', 
                 where=None, 
                 rowid: bool=False):

        import csv
        with open(self.filepath, newline='') as f:
            reader = csv.reader(f, delimiter=self.delimiter, quotechar=self.quotechar)
            next(reader)
            # iterate each line
            # for i in range(self.header_row):
            #     print('i', i)
            #     next(reader)
            for rownum, row in enumerate(reader, start=1):
                if len(self.columns) == len(row):
                    row_dict = dict(zip(self.columns, row))
                    # print('row_dict before where:', row_dict)
                    if columns is not None:
                        # print('columns arg', columns)
                        row_dict = {col: row_dict[col] for col in columns}
                        if row_type == 'list':
                            yield rownum, [row_dict[col] for col in columns]
                        else:
                            yield rownum, row_dict
                    else:
                        if row_type == 'list':
                            yield rownum, row
                        else:
                            yield rownum, row_dict
                else:
                    raise IndexError ('length of column and row is not the same at rownum {}. Possible issues were incorrect quoting or missing value.'.format(rownum))


    def set_to_sql_colmap(self, columns):
        if isinstance(columns, list) or isinstance(columns, tuple):
            return dict(zip(columns, columns))
        elif isinstance(columns, dict):
            return columns
            


import copy
from bintang.cell import Cell_Path_List
from bintang.row import Row_Table_Path
from bintang.log import log

def gen_tablepath(path_list, path_sep):
    pathl_norowid = [x for x in path_list if not isinstance(x, int)]
    if path_list[-1] == pathl_norowid[-1]:
        # this should work for pattern #1 and #2
        del pathl_norowid[-1]
        tablepath = path_sep.join(pathl_norowid)
        # if len(tablepath) > 2:
        #     tablepath = tablepath[1:]
        #log.debug(tablepath)
        return tablepath
    if (path_list[-2] == pathl_norowid[-1]) and isinstance(path_list[-1], int):
        # this shold work for pattern #3
        #log.debug(pathl_norowid)
        tablepath = path_sep.join(pathl_norowid)
        # if len(tablepath) > 2:
        #     tablepath = tablepath[1:]
        # #log.debug(tablepath)
        return tablepath


def gen_rowid(path_list):
    if isinstance(path_list[-1], str):
        pathl_copy = copy.deepcopy(path_list)
        del pathl_copy[-1]
        return ''.join(str(x) for x in pathl_copy)
    if isinstance(path_list[-1], int):
        return ''.join(str(x) for x in path_list)
    

def gen_table_path_row(path_list, path_sep, value):
    # there are two types of cell will be created as the result of data path location
    # 1. main cell
    # 2. parent cell (if any)
    # for the path_list 
    rowid = gen_rowid(path_list)
    row = Row_Table_Path(rowid)
    row.tablepath = gen_tablepath(path_list, path_sep)
    pathl = copy.deepcopy(path_list) 
    
    #log.debug('\n---------------------in gen_table_path_row (iterdict.py)--------------------------')
   
    # take the last item of the path
    if isinstance(pathl[-1], int):
        # if a list
        #log.debug('  a list...')
        del pathl[-1]
        del pathl[-1]
        #log.debug(f'  {len(pathl)}-> pathl after {pathl}')
    elif isinstance(pathl[-1], str):
        # if a dict
        #log.debug('  a dict...')
        del pathl[-1]
    
    # create 'parent key' cells
    while len(pathl) > 1: # if 0 then the root (/) will be included in the result set
        #log.debug(f'  {len(pathl)}->>> pathl {pathl}')
        #create the cell
        if isinstance(pathl[-1], int):
            # this is a list
            pathl_copy = copy.deepcopy(pathl) # to make it persistent and be used in making cell    
            cell = Cell_Path_List(len(pathl), pathl_copy, pathl[-1])
            cell.is_key = True
            row.add_cell(cell)
            #log.debug(f' len pathl {len(pathl)}-> cell {cell}')
            del pathl[-1]
            del pathl[-1]
            
        elif isinstance(pathl[-1], str):
            # this is a dict
            pathl_copy = copy.deepcopy(pathl) # to make it persistent and be used in making cell    
            cell = Cell_Path_List(len(pathl), pathl_copy, pathl[-1])
            cell.is_key = True
            row.add_cell(cell)
            #log.debug(f'  len pathl {len(pathl)}-> cell {cell}')
            del pathl[-1]

    # create the cell
    cell = Cell_Path_List(len(pathl), path_list, value)
    row.add_cell(cell)
    #log.debug(f'  len pathl {pathl}-> cell {cell}')
    #log.debug('---------------------out gen_table_path_row (iterdict.py)-----------------------') 
    return row           
    

def _iterdict(dict_or_list, path=['/']):
    if isinstance(dict_or_list, dict):
        iterator = dict_or_list.items()
    else:
        iterator = enumerate(dict_or_list)
    for k, v in iterator:
        yield path + [k], v
        if isinstance(v, (dict, list)):
            for k, v in _iterdict(v, path + [k]):
                yield k,v


def iterdict(dict_obj, root='/', path_sep='/', tablepaths=None):
    # yield a tprow (row with a table path)
    # also allow to yield only the row that matches tablepath list arg.
    for path_list, value in _iterdict(dict_obj, path=[root]): 
        #log.debug(f'k-> {path_list} v--> {value}')
        if isinstance(value,(list,dict)):
            continue # use continue and comment the below line value = None to create column with list/dict
            #value = '[]'  # include column with list/dict in the table but set value as None
        # print(path_list, value)
        if tablepaths is None: #len(tablepaths) == 0:
            # client wants to return all available tablepaths
            yield gen_table_path_row(path_list, path_sep, value)
        else: #elif len(tablepaths) > 0:
            # client wants to return specific tablepaths
            for tp in tablepaths:
                if tp == gen_tablepath(path_list, path_sep):
                    yield gen_table_path_row(path_list, path_sep, value)


import json
import bintang
from bintang.table_base import Base_Table
from bintang.column import Column
from bintang.cell import Cell
# from bintang.row import Row
from bintang.log import log
import types
from operator import itemgetter
from typing import Literal, Annotated, Any, get_origin
from itertools import product
from dataclasses import dataclass
import datetime

@dataclass
class ValueRange:
    lo: int
    hi: int

class Memory_Table(Base_Table):
    """Define a Bintang table object
       - provide columns to store a dictionary of column objects
       - provide rows to store a dictionary of row objects
    """
    def __init__(self,name, bing=None, conn=None):
        super().__init__(name, bing=bing)
        self.__columns = {}
        self.__rows = {}
        self.__temprows = []
        self.__last_assigned_columnid= 0 #
        # self.__last_assigned_rowid = 0 # for use when row created
        self.__last_assigned_idx = 0 # for use when add idx
        self.validate_funcs = {
            'int': self.validate_integer,
            'float': self.validate_float,
            'str': self.validate_string,
            'date': self.validate_date,
            'datetime': self.validate_datetime
            }
        
    def __getitem__(self, idx): # subscriptable version of self.get_row_asdict()
        #return self.__rows[idx] # bad design. a raw row isn't that useful at client code
        return self.get_row_asdict(idx)


    def __setitem__(self, rowcell, value): # subscriptable version of self.update_row()
        # where arg rowcell is a tuple of an idx of rows & column  passed by client code
        # client code signature: table_obj[idx,column] = value
        self.update_row(rowcell[0], rowcell[1], value)    
        

    def __repr__(self):
        tbl = {}
        tbl['name'] = self.name
        columns = []
        for k,v in self.__columns.items():
            columns.append(dict(id=v.id, name=v.name, ordinal_position=v.ordinal_position))
        columns.sort(key=lambda x: x['ordinal_position']) 
        tbl['columns'] = columns
        return json.dumps(tbl, indent=2)


    def __len__(self):
        return len(self.__rows)


    def get_size(self):
        return sys.getsizeof(self.__rows)
        
 
    def set_to_sql_colmap(self, columns):
        if isinstance(columns, list) or isinstance(columns, tuple):
            return dict(zip(columns, columns))
        elif isinstance(columns, dict):
            return columns

        
    # here the location of def_to_sql()    
    
    def _to_sql_upsert_dev(self, conn: str,
                    schema: str,
                    table: str,
                    columns: list[str],
                    on: list[tuple], # keys to match btw bintang and sql
                    method='prep') -> None:
        
        for row in self._iterrow_sql(conn):
            print(row)


    def _iterrow_sql(self, conn, sql_str=None, params=None):
        cursor = conn.cursor()
        if sql_str is None:
            sql_str = "SELECT * FROM {}".format(self.name)
        if params is not None:
            cursor.execute(sql_str, params)
        else:
            cursor.execute(sql_str)
        columns = [col[0] for col in cursor.description]
        for row in cursor.fetchall():
            yield dict(zip(columns, row))

    
    def gen_sql_stmt_create_table_dev(self, dbms, schema = None, str_size = None):
        # scanning table to get column properties
        # and assign the data type and size (when able)
        self.set_data_props()
        for col in self.get_columns():
            cobj = self.__columns[self.get_columnid(col)]
            if cobj.data_type is None:
                if len(cobj.data_props) == 0: # a column that has no data at all
                    cobj.data_type = 'str'  # force to str
                    cobj.column_size = str_size # force to str_size
                # loop through the data props for column that has it
                if 'str' in cobj.data_props:
                    cobj.data_type = 'str'
                    if str_size: # if user size to be used
                        if dbms == 'sqlserver':
                            if cobj.data_props['str']['column_size'] > 1:
                                cobj.column_size = str_size
                            else:
                                cobj.column_size = cobj.data_props['str']['column_size']
                    else:    
                        cobj.column_size = cobj.data_props['str']['column_size']
                    # loop through any other type if the column_size bigger, if it is assign it.
                    # for k, v in cobj.data_props.items():
                    #     if v['column_size'] > col_size:
                    #         col_size = v['column_size']
                    # cobj.column_size = col_size #cobj.data_props['str']['column_size']
                elif 'datetime' in cobj.data_props:
                    cobj.data_type = 'datetime'
                    cobj.column_size = cobj.data_props['datetime']['column_size']
                else: # just get the first one and break. to be observed later on!
                    for type, prop in cobj.data_props.items():
                        cobj.data_type = type
                        break
        
        # use type_map to translate the type
        create_columns = []
        for col in self.get_columns():
            cobj = self.__columns[self.get_columnid(col)]
            colname = cobj.name
            dtype = self.type_map[dbms]['type_mappings'][cobj.data_type]
            di_start = self.type_map[dbms]['delimited_identifiers']['start']
            di_end = self.type_map[dbms]['delimited_identifiers']['end']
            if cobj.data_type == 'str':
                col_size = cobj.column_size if cobj.column_size <= 4000 else 'max'
                create_item = [f'{di_start}{colname}{di_end}', f'{dtype}', f'({col_size})']
                create_columns.append(create_item)
            else:
                create_item = [f'{di_start}{colname}{di_end}', f'{dtype}']
                create_columns.append(create_item)       
        create_columns_str = []
        for i, citem in enumerate(create_columns):
            create_item_str = []
            if i == 0: # add tab if first, the rest will be added at later join
                create_item_str.append('\t')
            for i in citem:
                create_item_str.append(i)
            create_item_str.append('\n')
            create_columns_str.append(' '.join(create_item_str))
            schema_name = '' if schema is None else schema + '.'
        create_sqltable_templ = 'CREATE TABLE {}"{}" (\n{})'.format(schema_name, self.name, '\t,'.join(create_columns_str))
        return create_sqltable_templ 
    

    def add_column(self, name, data_type=None, column_size=None, min_value=None, max_value=None, min_length=None, max_length=None, required=False):
        # check if the passed name already exists
        columnid = self.get_columnid(name)
        if columnid is None:
            cobj = Column(name)
            if data_type is not None:
                cobj.data_type = data_type
            if column_size is not None:
                cobj.column_size = column_size
            if min_value is not None:
                cobj.min_value = min_value
            if max_value is not None:
                cobj.max_value = max_value
            if min_length is not None:
                cobj.min_length = min_length
            if max_length is not None:
                cobj.max_length = max_length
            if required is not None:
                cobj.required = required
            cobj.id = self.__last_assigned_columnid + 1
            cobj.ordinal_position = self.__last_assigned_columnid + 1
            self.__columns[cobj.id] = cobj
            self.__last_assigned_columnid= self.__last_assigned_columnid + 1
        else:
            log.debug(f'Warning! trying to add existing column "{name}".')
        

    def add_or_update_column(self, name, data_type=None, column_size=None, min_value=None, max_value=None, min_length=None, max_length=None, required=False):
        # check if the passed name already exists
        columnid = self.get_columnid(name)
        if columnid is None:
            self.add_column(name, data_type, column_size, min_value, max_value, min_length, max_length, required)
        else:
            self.update_column(name, data_type, column_size, None, min_value, max_value, min_length, max_length, required)


    def update_column(self,name, data_type=None, column_size=None, ordinal_position=None, min_value=None, max_value=None, min_length=None, max_length=None, required=False):
        # check if the passed name already exists
        columnid = self.get_columnid(name)
        if columnid is not None:
            if data_type is not None:
                self.__columns[columnid].data_type = data_type
            if column_size is not None:
                self.__columns[columnid].column_size = column_size
            if ordinal_position is not None:
                self.__columns[columnid].ordinal_position = ordinal_position
            if min_value is not None:
                self.__columns[columnid].min_value = min_value
            if max_value is not None:
                self.__columns[columnid].max_value = max_value
            if min_length is not None:
                self.__columns[columnid].min_length = min_length
            if max_length is not None:
                self.__columns[columnid].max_length = max_length
            if required is not None:
                self.__columns[columnid].required = required

                

    def get_columnid(self,column):
        for id, cobj in self.__columns.items():
            if cobj.get_name_uppercased() == column.upper():
                return id      

    
    def get_columnids(self,columns=None):
        columnids = []
        if columns is None: # assume user want all available column ids
            return [id for id in self.__columns.keys()]
        for column in columns:
            columnid = self.get_columnid(column)
            if columnid is None:
                raise ValueError('Cannot find column name {}.'.format(column))
            else:
                columnids.append(columnid)
        return columnids

            
    def rename_column(self,old_column,new_column):
        
        # for v in self.__columns.values():
        #     if v.name == old_column:
        #         v.name = new_column
        #         return
        old_column = self.validate_column(old_column) # validate user input
        columnid = self.get_columnid(old_column)
        if columnid is not None:
            self.__columns[columnid].name = new_column
       
        
    def drop_column(self,column):
        column = self.validate_column(column) # validate user input
        columnid = self.get_columnid(column)
        self.__columns.pop(columnid,None)
        

    def get_column(self,columnid):
        return self.__columns[columnid].name
    

    def order_columns(self, columns):
        # determine all columns
        columns_all = [x for x in columns]
        for col in self.get_columns():
            if col not in columns_all:
                columns_all.append(col)
        for i, col in enumerate(columns_all, 1):
            self.update_column(col, ordinal_position=i)


    # def _order_columns_sql(self, columns):
    #     # determine all columns
    #     columns_all = [x for x in columns]
    #     for col in self._get_columns_sql():
    #         if col not in columns_all:
    #             columns_all.append(col)
    #     for i, col in enumerate(columns_all, 10):
    #         self._update_column_sql(col, ordinal_position=i)        


    def get_columns(self) -> tuple:
        col_objs = [col for col in self.__columns.values()]
        col_objs.sort(key=lambda col: col.ordinal_position)
        sorted_columns = [col.name for col in col_objs]
        return tuple(sorted_columns)   


    def _get_columnnames_lced_XXX(self, columns=None) -> dict:
        # if self.type == 'SQLBACKEND':
        #     pass
        # elif self.type == 'FROMSQL':
        #     # execute sql
        #     # cursor = self.fromsql_conn.cursor()
        #     # cursor.execute(self.fromsql_str)
        #     # columns_fromsql = [col[0] for col in cursor.description]
        #     # return {x.name.lower(): x.name  for x in columns_fromsql}
        #     pass
        # else:
        #return {x.name.lower(): x.name  for x in self.__columns.values()}
        return {col.lower(): col  for col in self.get_columns()}


    def get_data_props(self, column):
        columnid = self.get_columnid(column)
        return self.__columns[columnid].data_props


    def get_column_object(self, column):
        columnid = self.get_columnid(column)
        return self.__columns[columnid]


    def get_column_object_byid(self,columnid):
        return self.__columns[columnid]    

 
    def validate_column(self, column):
        """return column as the one stored in table.columns"""
        if column.lower() in self._get_columnnames_lced().keys():
            return self._get_columnnames_lced().get(column.lower())
        else:
            similar_cols = bintang.core.get_similar_values(column, self.get_columns())
            if similar_cols:
                raise ValueError (f'Could not find column {repr(column)}. Did you mean {similar_cols}?')
            else:
                raise ValueError (f'Could not find column {repr(column)}')


    def validate_integer(self, value: int, cobj) -> tuple[int, bool, str | None]:
        #print(f"validate_integer called with value: {value} for column '{cobj.name}' with min_value={cobj.min_value} and max_value={cobj.max_value}")
        type_ = Annotated[int, ValueRange(cobj.min_value, cobj.max_value)]
        for annotation in type_.__metadata__:
            if isinstance(annotation, ValueRange):
                try:
                    value = int(float(value))  # Convert value to the base type float then integer
                    if cobj.min_value is not None and cobj.max_value is not None:
                        if not (cobj.min_value <= value <= cobj.max_value):
                            return (value, False, f"value is out of range [{cobj.min_value}, {cobj.max_value}]")
                    elif cobj.min_value is not None:
                        if not (value >= cobj.min_value):
                            return (value, False, f"value is below the minimum allowed value {cobj.min_value}")
                    elif cobj.max_value is not None:
                        if not (value <= cobj.max_value):
                            return (value, False, f"value is above the maximum allowed value {cobj.max_value}")
                    return (value, True, None) # return the converted value, validation result, and error message
                except (TypeError, IndexError, ValueError) as e:
                    return (value, False, f"ValueError: {e}")
    
    
    def validate_float(self, value: float, cobj) -> tuple[float,bool, str | None]:
        #print(f"validate_float called with value: {value} for column '{cobj.name}' with min_value={cobj.min_value} and max_value={cobj.max_value}")
        type_ = Annotated[float, ValueRange(cobj.min_value, cobj.max_value)]
        for annotation in type_.__metadata__:
            if isinstance(annotation, ValueRange):
                errors = []
                try:
                    value = float(value)  # Convert value to the base type float
                    if cobj.min_value is not None and cobj.max_value is not None:
                        if not (cobj.min_value <= value <= cobj.max_value):
                            return (value, False, f"value is out of range [{cobj.min_value}, {cobj.max_value}]")
                    elif cobj.min_value is not None:
                        if not (value >= cobj.min_value):
                            return (value, False, f"value is below the minimum allowed value {cobj.min_value}") 
                    elif cobj.max_value is not None:
                        if not (value <= cobj.max_value):
                            return (value, False, f"value is above the maximum allowed value {cobj.max_value}")
                    return (value, True, None) # return the converted value, validation result, and error message
                except (TypeError, IndexError, ValueError) as e:
                    return (value, False, f"ValueError: {e}")


    def validate_string(self, value: str, cobj) -> tuple[str, bool, str | None]:
        #print(f"validate_string called with value: '{value}' for column '{cobj.name}' with min_length={cobj.min_length} and max_length={cobj.max_length}")
        type_ = Annotated[str, ValueRange(cobj.min_length, cobj.max_length)]
        for annotation in type_.__metadata__:
            if isinstance(annotation, ValueRange):
                try:                    
                    value = str(value).strip()
                    if cobj.min_length is not None and cobj.max_length is not None:
                        if not (annotation.lo <= len(value) <= annotation.hi):
                            return (value, False, f"value has length {len(value)} which is out of allowed range [{annotation.lo}, {annotation.hi}]")
                    elif cobj.min_length is not None: 
                        if not (len(value) >= annotation.lo):
                            return (value, False, f"value has length {len(value)} which is below the minimum allowed length {annotation.lo}")
                    elif cobj.max_length is not None:
                        if not (len(value) <= annotation.hi):
                            return (value, False, f"value has length {len(value)} which is above the maximum allowed length {annotation.hi}")               
                    return (value, True, None) # return the converted value, validation result, and error message
                except (TypeError, IndexError, ValueError) as e:
                    return (value, False, f"column '{cobj.name}' has ValueError: {e}")
    

    def validate_date(self, value: datetime.date, cobj, parser) -> tuple[datetime.date, bool, str | None]:
        #print('validate_date called with value:', value)
        type_ = Annotated[datetime.date, ValueRange(cobj.min_value, cobj.max_value)]
        for annotation in type_.__metadata__:
            if isinstance(annotation, ValueRange):
                try: 
                    value = parser.parse(str(value)).date()  # Use dateutil.parser to parse the value into a date
                    if cobj.min_value is not None and cobj.max_value is not None:
                        if not (annotation.lo <= value <= annotation.hi):
                            return (value, False, f"value is out of range [{annotation.lo}, {annotation.hi}]")
                    elif cobj.min_value is not None:
                        if not (value >= annotation.lo):
                            return (value, False, f"value is below the minimum allowed value {annotation.lo}")
                    elif cobj.max_value is not None:
                        if not (value <= annotation.hi):
                            return (value, False, f"value is above the maximum allowed value {annotation.hi}")
                    return (value, True, None) # return the converted value, validation result, and error message
                except (TypeError, IndexError, ValueError) as e:
                    return (value, False, f"ValueError: {e}")
    

    def validate_datetime(self, value: datetime.datetime, cobj, parser) -> tuple[datetime.datetime, bool, str | None]:
        #print('validate_datetime called with value:', value)
        type_ = Annotated[datetime.datetime, ValueRange(cobj.min_value, cobj.max_value)]
        for annotation in type_.__metadata__:
            if isinstance(annotation, ValueRange):
                try:                    
                    value = parser.parse(str(value))  # Use dateutil.parser to parse the value into a datetime
                    if cobj.min_value is not None and cobj.max_value is not None:
                        if not (annotation.lo <= value <= annotation.hi):
                            return (value, False, f"value is out of range [{annotation.lo}, {annotation.hi}]")
                    elif cobj.min_value is not None:
                        if not (value >= annotation.lo):
                            return (value, False, f"value is below the minimum allowed value {annotation.lo}")
                    elif cobj.max_value is not None:
                        if not (value <= annotation.hi):
                            return (value, False, f"value is above the maximum allowed value {annotation.hi}") 
                    return (value, True, None) # return the converted value, validation result, and error message
                except (TypeError, IndexError, ValueError) as e:
                    return (value, False, f"ValueError: {e}")


    def valrows(self):
        """ validate every column that has data_type assigned then
        put the results to a new table with name into.
        any passing validation will be inserted into the new table and any failed validation will be logged with error message and put into the new table as well.
        """
        try:
            from dateutil import parser
        except ImportError as e:
            print(e)
            raise ImportError("valrows() require python-dateutil. please install it for eg. pip install python-dateutil")

        # get columns that have data type assigned.
        data_typed_columnids = []
        for colid in self.get_columnids():
            if self.__columns[colid].data_type is not None:
                data_typed_columnids.append(colid)      
        # iterate through rows and validate each cell value with its column properties
        # only validate a non None value. if None value then only check if it is requred.
        for idx, row in self._iterrows():
            results = []
            for colid in data_typed_columnids:
                cobj = self.__columns[colid]
                value = None
                if colid in row.cells: # check if the cell exists
                    value = row.cells[colid].value 
                    if isinstance(value, str) and value == '':
                        value = None 
                        res = (colid, None, True, None) # force any '' to None
                if value is None:
                    if self.__columns[colid].required:
                        res = (colid, None, False, f"column is required but value is None")
                    # else: # comment to avoid handling later
                    #     res = (colid, None, True, None)
                elif value is not None:
                    if cobj.data_type in ['date','datetime']:
                        res = (colid,) + self.validate_funcs[cobj.data_type](value, self.__columns[colid], parser)
                    else:
                        res = (colid,) + self.validate_funcs[cobj.data_type](value, self.__columns[colid])
                results.append(res)
            yield idx, tuple(results)


    def extract_invalid_msg_(self,results):
        res_msg = []
        for res in results:
            if not res[2]:
                item= f"column {self.get_column(res[0])} - {res[3]}"
                res_msg.append(item)      
        return '; '.join(res_msg)


    def validate(self, into_invalid_table = None):
        """ validate the table and ouput valid and invalid table 
        """
        # iterate through rows and validate each cell value with its column properties
        # then any passing validation will be inserted into the new table and any failed validation will be logged with error message and put into the new table as well.
        invalids = [] # to track invalid rows
        if into_invalid_table:
            invalid_tobj = self.bing.create_table(into_invalid_table)
        for idx, results in self.valrows():
            # print(idx,  results)
            if all(x[2] for x in results): # if valid update existing row and insert into valid_tobj
                row = self.__rows[idx]
                # update row for each valid column in the results
                for res in results:
                    print(res)
                    self.__rows[idx].cells[res[0]].value = res[1]
            else:
                invalids.append(idx)
                if into_invalid_table:
                    invalid_records = self.get_row_asdict(idx)
                    invalid_records['invalid_idx'] = idx
                    invalid_records['error_msg'] = self.extract_invalid_msg_(results)
                    invalid_tobj.insert(invalid_records, index=idx)

        # delete invalid row out existing table
        for idx in invalids:
            del self.__rows[idx]


    def copy_index(self, column='idx',at_start=False):
        for idx in self.__rows: 
            if isinstance(idx, tuple): # if index generated from a groupby
                self.update_row(idx, column, str(idx))
            else:
                self.update_row(idx, column, idx)
        if at_start:
            self.order_columns([column])


    def get_index(self, condition_column, condition_value):
        """will return the first index which value matches the condition"""
        for idx, row in self.iterrows():
            # test if string
            # if type(row[condition_columnname])==str:pass
            if row[condition_column] == condition_value:
                return idx

    def get_indexes(self, condition_column, condition_value):
        """will return a list of indexes which value matches the condition"""
        indexes = []
        for idx, row in self.iterrows():
            # test if string
            # if type(row[condition_columnname])==str:pass
            if row[condition_column] == condition_value:
                indexes.append(idx)
        return indexes


    def get_all_indexes(self):
        """will return a list of all indexes"""
        return tuple(self.__rows.keys())


    # moved def get_value() to base table.


    def index_exists(self, index):
        if index in self.__rows:
            return True
        else:
            return False


    #remove def make_row to base table


    def insert(self, dict_or_columns, values=None, index=None):
        """ restrict arguments for record insertion for this function as the followings:
        1. a dictionary pass to record param
        2. list values pass to record param and list columns pass to column param.
        """
        if isinstance(dict_or_columns, dict):
            row = self.make_row()
            for idx, (col, val) in enumerate(dict_or_columns.items()):
                cell = self.make_cell(col,val)
                row.add_cell(cell) # add to row
            self.add_row(row, index)   
            
                                                  
        elif isinstance(dict_or_columns,list) or isinstance(dict_or_columns,tuple) or isinstance(values,list) or isinstance(values,tuple):
            row = self.make_row()
            for idx, col in enumerate(dict_or_columns):
                cell = self.make_cell(col,values[idx])
                row.add_cell(cell) # add to rows
            self.add_row(row, index) 
        else:
            raise ValueError("Arg for record set for dictionary or list/tuple of values with list/tuple of columns.")
        

    def _insert(self, record, columns=None, index=None):
        """ restrict arguments for record insertion for this function as the followings:
        1. a pair of columns and its single-row values
        deprecated2. a pair of columns and its multiple-row values (as a list of tuple)
        """
        if isinstance(record, dict):
            row = self.make_row()
            for idx, (col, val) in enumerate(record.items()):
                cell = self.make_cell(col,val)
                row.add_cell(cell) # add to row
            self.add_row(row, index)                                    
        elif isinstance(columns,list) or isinstance(columns,tuple) or isinstance(record,list) or isinstance(record,tuple):
            row = self.make_row()
            for idx, col in enumerate(columns):
                cell = self.make_cell(col,record[idx])
                row.add_cell(cell) # add to rows
            if self.__be is not None:
                self.__temprows.append(json.dumps({v.columnid: v.value for v in row.cells.values()}))
                if len(self.__temprows) == self.__be.MAX_ROW_SQL_INSERT:
                    self.add_row_into_be()
            else:
                self.add_row(row, index)    
        else:
            raise ValueError("Arg for record set for dictionary or list/tuple of values with list/tuple of columns.")    
            
            
    def add_row_into_be(self):
        self.__be.add_row(self.name, self.__temprows)
        self.__temprows.clear()      


    def _insert_dict(self, adict):
        self.insert([x for x in adict.keys()], [x for x in adict.values()])


    def insert_todb(self,columns,values):
        """ restrict arguments for data insertion for this function as the followings:
        1. a pair of columns and its single-row values
        2. a pair of columns and its multiple-row values (as a list of tuple)
        """
        if isinstance(columns,list) and not isinstance(values[0],tuple):
            row = self.make_row()
            for idx, column in enumerate(columns):
                cell = self.make_cell(column,values[idx])
                row.add_cell(cell)
            # add to rows
            self.add_row(row)
        else:
            raise ValueError("Insert only allows a pair of columns and values in a list or tuple")


    def get_rowidx_byrowid(self, rowid):
        debug = False
        # if debug:
        #     print("\n  ------------------in get_row_rowid (table.py) --------------------")
        for idx, row in self.__rows.items():
            if debug:
                print(idx, row)
            if row.id == rowid:
                return idx
        # if debug:
        #     print("\n  ------------------out get_row_rowid (table.py) -------------------")
        return None    


    def upsert_table_path_row_moved(self, tprow):
        #log.debug("\n  ------------------in upsert_table_path_row (table.py) --------------------")
        # extract the rowid and use it as the table index (the key of rows{})
        # create a row if the rowid not found in the table's index
        res_idx = self.get_rowidx_byrowid(tprow.id)
        
        if res_idx is None:
            #log.debug(f'inserting... row does not exist {tprow}')
            row = self.make_row(tprow.id)
            # re-make cells from tprow
            for id, c in tprow.cells.items():
                #log.debug(f'{id} cell: {c}')
                if c.is_key == True:
                    cell = self.make_cell(self.PARENT_PREFIX + c.get_column(), c.value)
                else:
                    cell = self.make_cell(c.get_column(), c.value)

                row.add_cell(cell)
            # add to rows
            self.add_row(row)
            
        elif res_idx is not None:
            #log.debug(f"updating... row exists {tprow}")
            for id, c in tprow.cells.items():
                if c.is_key == True:
                    cell = self.make_cell(self.PARENT_PREFIX + c.get_column(), c.value)
                else:
                    cell = self.make_cell(c.get_column(), c.value)
                self.__rows[res_idx].add_cell(cell)        
        #log.debug("\n  ------------------out upsert_table_path_row (table.py)-------------------")


    def make_cell(self,column,value,new_column=True):
        if not isinstance(column, str):
            raise ValueError("column name must be a string type!")
        columnid = self.get_columnid(column)
        if columnid is None: # if columnid is None then assume user wants a new column
            if new_column == True:
                self.add_column(column)
                columnid = self.get_columnid(column) # reassign the columnid
        # if columnid is None:
        #     raise ValueError("Cannot make cell due to None column name.")    
        return Cell(columnid,value)


    def add_row(self, row, index=None):
        if index is None:
            index = self.__last_assigned_idx + 1
            self.__last_assigned_idx += 1
        self.__rows[index] = row      


    def delete(self, where):
        # get all index
        indexes = [x for x in self.__rows]
        for idx in indexes:
            try:
                if where(self.get_row_asdict(idx)):
                    self.delete_row(idx)
            except:
                pass


    def delete_row(self,index):
        self.__rows.pop(index,None)


    def delete_rows(self,indexes):
        for index in indexes:
            self.delete_row(index)


    def _get_row_ascells(self,idx):
        return self.__rows[idx]


    def get_valid_columnname(self, column):
        columnid = self.get_columnid(column)   # column in this line passed by user
        column = self.get_column(columnid) # ensure the same column passed as result.
        return column


    def get_row(self, index, columns=None, rowid=False, row_type='dict'):
        if row_type.lower()=='list':
            return self.get_row_aslist(index, columns)
        else:
            return self.get_row_asdict(index, columns, rowid)

    def get_row_(self, index):
        return  self.__rows[index]       
           

    def get_row_asdict(self, idx, columns=None, rowid=False):
        if idx in self.__rows:
            if columns is not None:
                columns = self.validate_columns(columns)
            else:
                columns = self.get_columns()
            columnids = self.get_columnids(self.get_columns())    
            return self._gen_row_asdict(self.__rows[idx],columnids, rowid)


    def _get_row_asdict(self, idx, columnids=None, rowid=False):
        """ internal function to get row as dict
            no need to check if idx exists in self.__rows 
        """
        if columnids is None:
            columnids = self.get_columnids()
        return self._gen_row_asdict(self.__rows[idx],columnids, rowid)        


    def _get_row_sql(self, idx, columns=None, row_type='dict'):
        cursor = self.conn.cursor()
        sql = 'SELECT cells from {} WHERE idx=?'.format(self.name)
        params = [idx]
        cursor = self.conn.cursor()
        cursor.execute(sql, params)
        res = cursor.fetchone()
        if res:
            if columns is None:
                columns = self._get_columns_sql()
            if row_type.lower() == 'list':
                return [x for x in self._gen_row_dict_sql(res['cells'], columns).values()]
            else:
                return self._gen_row_dict_sql(res['cells'], columns)
            

    def _gen_row_asdict(self, row, columnids, rowid=False):
        """ Generate a row as a dictionary.
        Args:
            row: a Row object
            columns: a list of column names
            rowid: if True then add rowid_ as the first key in the result dict
        Returns:
            a dictionary with column names as keys and cell values as values.
        """
        res = {}
        # if rowid == True:
        #     res['rowid_'] = row.id # add rowid for internal purpose eg. a merged table
        for columnid in columnids:
            column = self.get_column(columnid) # ensure the same column passed as result.
            res[column] = row.get_value(columnid)     
        return res


    def get_row_aslist(self, idx, columns=None):
        if columns is None:
            columns = self.get_columns()
        columnids = self.get_columnids(columns)
        return self._gen_row_aslist(self.__rows[idx],columnids)


    def _gen_row_aslist(self, row, columnids):
        return row.get_values(columnids)


    def _normalise_row(self, row, columnids, rowid=False):
        # we'll add a new cell as None if never existed when reading the source
        for cid in columnids:
            if cid not in row.cells:
                cell = Cell(cid,None)  # a none cell or never existed for the row at the source
                row.add_cell(cell)
        return row  


    def iterrows(self, 
                 columns: list=None, 
                 row_type: Literal['dict','list']='dict', 
                 where=None, 
                 rowid: bool=False):
        
        # validate user's args
        if columns is not None:
            columns = self.validate_columns(columns)
        else:
            columns = self.get_columns() # assign all available column names
        columnids = self.get_columnids(columns)    
        if row_type == 'list':
            if where is not None:
                for idx, row in self.__rows.items():
                    if where(self._gen_row_asdict(row,columnids,rowid)):
                        yield idx, self._gen_row_aslist(row,columnids)
            else:
                for idx, row in self.__rows.items():
                    yield idx, self._gen_row_aslist(row,columnids)
        elif row_type == 'object':
            for idx, row in self.__rows.items():
                yield idx, self._normalise_row(row, columnids)


        else: # assume row_type is dict
            if where is not None:
                for idx, row in self.__rows.items():
                    if where(self._gen_row_asdict(row, columnids, rowid)):
                        yield idx, self._gen_row_asdict(row,columnids,rowid)
            else:
                for idx, row in self.__rows.items():
                    yield idx, self._gen_row_asdict(row,columnids,rowid)
      
    
    def _iterrows(self):
        for idx, row in self.__rows.items():
            yield idx, row
    
    
    def set_data_props(self):
        """ scan table to obtain columns properties - data type, column size (if str type then the max of len of string)"""
        columnids = self.get_columnids()
        columns = self.get_columns()
        for idx, row in self.__rows.items():
            for columnid in columnids:
                # log.debug(row)
                # log.debug('id: {}, value: {}'.format(idx , self.__columns[columnid].name))
                if columnid in row.cells:
                    self.set_data_props_datatype(columnid, row.cells[columnid].value)
                    if row.cells[columnid].value is not None:
                        if type(row.cells[columnid].value).__name__ == 'str':
                            self.set_data_props_str_column_size(columnid, row.cells[columnid].value)
                        if type(row.cells[columnid].value).__name__ not in ['str','int','float','bool','datetime']:
                            self.set_data_props_other_column_size(columnid, row.cells[columnid].value)


    def set_data_props_datatype(self, columnid,value):
        # set a column's data_type
        # set predefined column size for known data type, int, float, bool, datetime
        # while for str and the rest, need to go through all column cells and get greatest length.
        if value is not None:
            if type(value).__name__ not in self.__columns[columnid].data_props:
                self.__columns[columnid].data_props[type(value).__name__] = dict(column_size=1)
                if type(value).__name__ == 'int':
                    self.__columns[columnid].data_props['int']['column_size'] = 10
                if type(value).__name__ == 'float':
                    self.__columns[columnid].data_props['float']['column_size'] = 15
                if type(value).__name__ == 'bool':
                    self.__columns[columnid].data_props['bool']['column_size'] = 1
                if type(value).__name__ == 'datetime':
                    self.__columns[columnid].data_props['datetime']['column_size'] = 19        

                
    def set_data_props_str_column_size(self, columnid, value):
        if len(value) > self.__columns[columnid].data_props['str']['column_size']:
            self.__columns[columnid].data_props['str']['column_size'] = len(value)


    def set_data_props_other_column_size(self, columnid, value):
        if len(str(value)) > self.__columns[columnid].data_props[type(value).__name__]['column_size']:
            self.__columns[columnid].data_props[type(value).__name__]['column_size'] = len(str(value))         


    def print_columns_info(self):
        self.set_data_props()
        dp_tobj = Memory_Table(f'{self.name} - Columns Info')
        row_dict = {}
        for col in self.get_columns():
            cobj = self.get_column_object(col)
            row_dict['id'] = cobj.id
            row_dict['name'] = cobj.name
            row_dict['ordinal_position'] = cobj.ordinal_position
            row_dict['data_type'] = cobj.data_type
            row_dict['column_size'] = cobj.column_size
            row_dict['decimal_digits'] = cobj.decimal_digits
            row_dict['data_props'] = cobj.data_props
            dp_tobj.insert(row_dict)
        dp_tobj.print()

    
    
    def _get_index_column_size(self):
        column_size = 0
        for idx in self.__rows:
            # print(idx)
            # print(len(''.join(idx)))
            if len(str(idx)) > column_size:
                column_size = len(str(idx))
        return max(len('idx'),column_size)
    

    def _gen_print_line(self, length, col_div_pos, square=False):
        line_chars = []
        if square:
            for x in range(length):
                if x in [0,length-1]:
                    line_chars.append('+')
                elif x  in col_div_pos:
                    line_chars.append('+')
                else:
                    line_chars.append('-')
        else:
            for x in range(length):
                if x  in col_div_pos:
                    line_chars.append('+')
                else:
                    line_chars.append('-')
        return ''.join(line_chars)

    def _gen_heading_col_str(self, heading_col, square=False):
        # heading_col_str = '|'.join(heading_col)
        if square:
            temp_heading_col = list('|'.join(heading_col))
            temp_heading_col[0] = '|'
            temp_heading_col[-1] = '|'
            return ''.join(temp_heading_col)
        else:
            return '|'.join(heading_col)

    def _gen_print_padded_cells_str(self, padded_cells, square=False):
        if square:
            temp_cells = list('|'.join(padded_cells))
            temp_cells[0] = '|'
            temp_cells[-1] = '|'
            # print(temp_cells)
            # quit()
            return ''.join(temp_cells)
        else:
            return '|'.join(padded_cells)


        

    def print(self, columns=None, show_data_type=False, square=False, limit = 10):
        # validate columns arg
        if columns: #user want specific columns
            columns_ = self.validate_columns(columns)
        else:
            columns_ = self.get_columns()
            
        PAD_SIZE = 4
        self.set_data_props() # generate data based type and column_size
        # generate heading_col string
        header_data_types = []
        heading_col = []
        col_div_pos = [] #[idx_max_width]
        for col in columns_: #self.get_columns():
            cobj = self.get_column_object(col)
            _data_types_str = ','.join(cobj.get_data_types())
            dp = cobj.get_greatest_column_size_data_prop()
            max_width = PAD_SIZE # initial value
            if dp is None:
                max_width = max(4, len(col)) + PAD_SIZE
            else:
                max_width = max(next(iter(dp.values()))['column_size'],len(col)) + PAD_SIZE
            if len(col_div_pos)==0:
                col_div_pos.append(max_width)
            else:    
                col_divider_pos = col_div_pos[-1] + max_width + 1
                col_div_pos.append(col_divider_pos)

            if show_data_type:
                header_dt = _data_types_str.center(max_width)
                header_data_types.append(header_dt)
            header_col = col.center(max_width)
            heading_col.append(header_col)

        if show_data_type:
            header_data_types_str = '|'.join(header_data_types)
        heading_col_str = self._gen_heading_col_str(heading_col, square=square)
  
        print(('Table: ' + self.name).center(len(heading_col_str))) # print title
        if show_data_type:
            print(header_data_types_str) # print header data type
        print(self._gen_print_line(len(heading_col_str), col_div_pos,  square=square))
        print(heading_col_str) # print heading_col str
        print(self._gen_print_line(len(heading_col_str), col_div_pos,  square=square))
        
        # generate row string
        for idx, row in self.iterrows(columns_):
            padded_cells = []
            for col, val in row.items():
                cobj = self.get_column_object(col)
                dp = cobj.get_greatest_column_size_data_prop()
                max_width = PAD_SIZE # initial value
                if dp is None:
                    max_width = max(4, len(col)) + PAD_SIZE
                else:
                    max_width = max(next(iter(dp.values()))['column_size'],len(col)) + PAD_SIZE
                # if type(val).__name__ not in ['str','int','float','bool','datetime','NoneType']:
                #     padded_cells.append((' ' + str(type(val).__name__) + ' obj').rjust(max_width))
                if cobj.has_string_data_type():
                   padded_cells.append(('  ' + str(val)).ljust(max_width)) #add double spaces for readable
                else:
                    padded_cells.append((str(val) + '  ').rjust(max_width)) # add double spaces for readable
            #padded_cells_str = '|'.join(padded_cells)
            print(self._gen_print_padded_cells_str(padded_cells,  square=square))
            if idx == limit: break
        print(self._gen_print_line(len(heading_col_str), col_div_pos,  square=square))
        print('({} rows)'.format(len(self)))    

    
    def update_row(self,idx,column,value):
        if idx in self.__rows:
            self.__rows[idx].add_cell(self.make_cell(column,value))
        else:
            fn = inspect.stack()[0][3]
            warnings.warn(f'Warning! {fn}() trying to update a non existed row (ie. idx {idx}).')

   
    def update_row_all(self,column,value):
        for idx in self.__rows:
            self.__rows[idx].add_cell(self.make_cell(column,value))    


    def exists(self, where):
        for idx, row in self.iterrows():
            try:
                if where(row):
                    return True
            except:
                print(row)
                pass
        return False    

    
    def get_index_lambda(self,expr):
        for idx, row in self.iterrows():
            log.debug('idx {} row{}:'.format(idx, row))
            ret = expr(row)
            if ret:
                return idx


    def get_indexes_lambda(self,expr):
        indexes = []
        for idx, row in self.iterrows():
            #log.debug('idx {} row{}:'.format(idx, row))
            ret = expr(row)
            if ret:
                indexes.append(idx)       
        return indexes


    def filter(self, 
                columns: list=None, 
                where: callable=None):
        tobj = Memory_Table('filtered')
        if columns is None: # all columns
            columns = self.get_columns()
        else:
            columns = self.validate_columns(columns)
        for idx, row in self.iterrows(columns):
            if where:
                try:
                    if where(row):
                        tobj.insert(row)
                except Exception as e:
                    pass
                    # log.debug(f'false at {idx}')
                    # log.warning(e)    
            else:
                tobj.insert(row)
        return tobj


    def innerjoin(self
                ,rtable # either string or Table object
                ,on: list[tuple] # list of lkey & r key tuple
                ,into: str
                ,out_lcolumns: list=None
                ,out_rcolumns: list=None
                ,rowid=False):
        
        rtable_obj = None
        if isinstance(rtable,Memory_Table):
            rtable_obj = rtable
            rtable = rtable.name
        else:
            rtable_obj = self.bing[rtable]
        
        # validate input eg. column etc
        lkeys = [x[0] for x in on] # generate lkeys from on (1st sequence)
        rkeys = [x[1] for x in on] # generate rkeys from on (2nd sequence)
        lkeys = self.validate_columns(lkeys)
        #rkeys = self[rtable].validate_columns(rkeys)
        rkeys = rtable_obj.validate_columns(rkeys)
        if out_lcolumns is not None:
            out_lcolumns = self.validate_columns(out_lcolumns)
        if out_rcolumns is not None:
            # out_rcolumns = self[rtable].validate_columns(out_rcolumns)
            out_rcolumns = rtable_obj.validate_columns(out_rcolumns)

        # resolve columns conflicts
        rcol_resolved = self._resolve_join_columns(rtable_obj, rowid)
        # create an output table
        out_table = into
        out_tobj = self.bing.create_table(out_table)
        out_tobj = self.bing.get_table(out_table)
        
        # for debuging create merged table to store the matching rowids
        #merged = self.create_table("merged",["lrowid","rrowid"])

        numof_keys = len(on) #(lkeys)
        # loop left table
        for lidx, lrow in self.iterrows(columns=lkeys, rowid=rowid):
            # loop right table
            for ridx, rrow in rtable_obj.iterrows(columns=rkeys, rowid=rowid):
                matches = 0 # store matches for each rrow
                # compare value for any matching keys, if TRUE then increment matches
                for i in range(numof_keys): 
                    if bintang.match(lrow[lkeys[i]], rrow[rkeys[i]]):
                        matches += 1 # incremented!
                if matches == numof_keys: # if fully matched, create the row & add into the output table
                    #debug merged.insert(["lrowid","rrowid"], [lrow["_rowid"], rrow["_rowid"]])
                    #and add the row to output table
                    outrow = out_tobj.make_row()
                    # add cells from left table
                    log.debug(f'{lidx}')
                    outrow = self._add_lcell(lidx,  outrow, out_table, out_lcolumns, rowid)
                    # add cells from right table
                    outrow = self._add_rcell(ridx, rtable_obj, outrow, out_table, out_rcolumns, rcol_resolved, rowid)   
                    out_tobj.add_row(outrow)
        #debug merged.print() 
        return out_tobj


    def _resolve_join_columns (self, rtable_obj, rowid=False):
        # resolve any conflict column by prefixing its tablename
        # notes: conflict column occurs when the same column being used in two joining table.
        rcol_resolved = {}
        if rowid:
            rcol_resolved['rowid_'] = rtable_obj.name + '_' + 'rowid_'
        lcolumns = self.get_columns()
        for col in rtable_obj.get_columns():
            if col in lcolumns:
                rcol_resolved[col] = rtable_obj.name + '_' + col
            else:
                rcol_resolved[col] = col
        return rcol_resolved
    
    def _add_lcell(self, lidx, outrow, out_table, out_lcolumns, rowid):
        for k, v in self.get_row_asdict(lidx,rowid=rowid).items():
            if out_lcolumns is None:
                # incude all columns
                cell = self.bing[out_table].make_cell(k,v)
                outrow.add_cell(cell)
            if out_lcolumns is not None:
                # include only the passed columns
                if k in out_lcolumns:
                    cell = self.bing[out_table].make_cell(k,v)
                    outrow.add_cell(cell)
        return outrow            


    def _add_rcell(self, ridx, rtable, outrow, out_table, out_rcolumns, rcol_resolved, rowid):
        for k, v in rtable.get_row_asdict(ridx,rowid=rowid).items():
            if out_rcolumns is None:
                cell = self.bing[out_table].make_cell(rcol_resolved[k],v)
                outrow.add_cell(cell)
            if out_rcolumns is not None:
                # include only the passed columns
                if k in out_rcolumns:
                    cell = self.bing[out_table].make_cell(rcol_resolved[k],v)
                    outrow.add_cell(cell)
        return outrow
  

    def blookup(self, 
                lkp_table: object, 
                on: list[tuple], 
                ret_columns: list[str] | list[tuple]):
        lkp_table_obj = None
        if isinstance(lkp_table,Memory_Table):
            lkp_table_obj = lkp_table
            lkp_table = lkp_table.name
        else:
            lkp_table_obj = self.bing[lkp_table]
        
        # validate input eg. column etc
        lkeys = [x[0] for x in on] # generate lkeys from on (1st sequence)
        rkeys = [x[1] for x in on] # generate rkeys from on (2nd sequence)
        lkeys = self.validate_columns(lkeys)
        rkeys = lkp_table_obj.validate_columns(rkeys)
        lcolumn_prematch = self.get_columns() # will use this list when matching occurs below
        # validate ret_columns once:
        # otherwise will make lots of call later
        valid_ret_columns = []
        for col in ret_columns:
            if isinstance(col,str):
                col = lkp_table_obj.validate_column(col)
                valid_ret_columns.append(col)
            if isinstance(col, tuple):
                valid_column  = lkp_table_obj.validate_column(col[0])
                col_ = tuple([valid_column,col[1]])
                valid_ret_columns.append(col_)
        # add these valid_ret_column to the left table
        for col in valid_ret_columns:
            if isinstance(col, str):
                self.add_column(col)
            if isinstance(col, tuple):
                self.add_column(col[1]) # use the alias
        # for each matching, update it
        for lidx, results in self.cmprows(lkp_table, on=on, find_all=False):
            # update this table lrow for each ret_columns
            for col in valid_ret_columns:
                if isinstance(col, str): # if col is a column
                    # extract value at the first value of the first tuple
                    value = lkp_table_obj._get_row_ascells(results[0][0]).get_value(lkp_table_obj.get_columnid(col))
                    if col in lcolumn_prematch:
                        # update left table with the lkp_table name as prefix
                        # this is to avoid column name conflict
                        self.update_row(lidx, self.bing[lkp_table].name + '_' + col, value)
                    else:
                        self.update_row(lidx, col, value)
                if isinstance(col, tuple): # if a tuple (0=column to return from lkp_table, 1=as_column)
                    value = lkp_table_obj._get_row(results[0][0]).get_value(lkp_table_obj.get_columnid(col[0]))
                    self.update_row(lidx, col[1], value)
    

    # cmprows() moved to base table


    def _scanfuzzy(self, lkeys, lkp_table_obj, rkeys, min_ratios):
        """ will yield two items. a left idx and result (a list of sorted matched right index and ratio tuple)
        for eg. yield 1, [(4, [80, 90]),(3, [70, 69])]
        where: 1 = the left index, 
               [(4, [80, 90]),(3, [70, 69])] = the sorted tuple list
        """
        lenof_keys = len(lkeys)
        for lidx, lrow in self.iterrows(lkeys, rowid=True):
            res_dict = {}
            for ridx, rrow in lkp_table_obj.iterrows():
                matches = 0
                ratios = []
                for i in range(lenof_keys):
                    ratio = bintang.get_fuzzy_ratio(lrow[lkeys[i]], rrow[rkeys[i]])
                    if ratio >= min_ratios[i]:
                        matches += 1 
                        ratios.append(ratio)
                if matches == lenof_keys:
                    res_dict[ridx] = ratios
            
            if len(res_dict) > 1:
                res_tuples_sorted = sorted(res_dict.items(), key=itemgetter(1), reverse=True)
                yield lidx, res_tuples_sorted
            if len(res_dict) == 1:
                yield lidx, list(res_dict.items())


    def fuzzy_cmprows(self, lkp_table, on: list[tuple]=None, min_ratio=0.70, min_matches=1, find_all=True):
        """ 
        will yield two items. a left idx and result with the following format and example:
        lidx, [ridx,(lcol1,rcol1,score1),(lcol2,rcol2,score2),...]
        1 [(2, ((11, 11, 0.75), (12, 12, 1.0)))]
        result is sorted based on the sum of ratios from highest to lowest.

        args:
        lkp_table: the lookup table that we compare against
        on: a list of matching column tuples that also includes min score. If not provided then all possible columns matching (cartasian product) will be carried out.
        find_all (bool): if True (default) then check all the lookup data, while  If False then exit on first match
        """
        on_ = [] # will hold column id keys instead column name
        if on:
            req_matches = len(on)
            for tup in on:
                # extract (lcol, rcol, ratio) from the tuple
                tup_ = (self.get_columnid(tup[0]), self.bing[lkp_table].get_columnid(tup[1]), tup[2])
                on_.append(tup_)
        else:
            req_matches = min_matches
            
            lcolumns = self.get_columns()
            rcolumns = self.bing[lkp_table].get_columns()
            ratios = [min_ratio] * len(lcolumns)
            # establish the on
            on = list(product(lcolumns, rcolumns))
            for tup in on:
                # extract (lcol, rcol, ratio) from the tuple
                tup_ = (self.get_columnid(tup[0]), self.bing[lkp_table].get_columnid(tup[1]), min_ratio)
                on_.append(tup_)          
        for lidx, lrow in self._iterrows():
            results = []
            for ridx, rrow in self.bing[lkp_table]._iterrows():
                matches = 0
                col_results = []
                for i in range(len(on_)):
                    if on_[i][0] in lrow.cells and on_[i][1] in rrow.cells: # check if both cells exist
                        ratio = bintang.get_fuzzy_ratio(str(lrow.cells[on_[i][0]].value), str(rrow.cells[on_[i][1]].value))
                        if ratio >= on_[i][2]:
                            matches += 1 
                            cmp = (on_[i][0], on_[i][1], ratio) # (lcol, rcol, score)
                            col_results.append(cmp)
                if matches >= req_matches:
                    results.append((ridx, tuple(col_results)))
                    if not find_all:
                        break
            if len(results) > 1: # then sort the results
                results = sorted(
                                results, 
                                key=lambda x: sum(colres[2] for colres in x[1]), # x=row_results, colres=col_results
                                reverse=True
                            )
                yield lidx, results
            elif len(results) == 1:
                yield lidx, results
                

    def fmatch(self,
                lkp_table: object,
                on: list[tuple],
                into: str,
                # ret_matches: bool = False
                ):
        """ Perform fuzzy matching between two tables.
            lkp_table: the lookup table to match with.
            on: a list of tuples, each tuple contains (left column, right column, min_ratio).
                eg. [('last_name', 'surname', 80), ('address', 'address', 70)]
            into: the name of the output table to hold the results.
        """
        lkp_table_obj = None
        if isinstance(lkp_table,Memory_Table):
            lkp_table_obj = lkp_table
            lkp_table = lkp_table.name
        else:
            lkp_table_obj = self.bing[lkp_table]
        
        # validate input eg. column etc
        lkeys = [x[0] for x in on] # generate lkeys from on (1st sequence)
        rkeys = [x[1] for x in on] # generate rkeys from on (2nd sequence)
        min_ratios = [x[2] for x in on] # generate ratios from on (3rd sequence)
        lkeys = self.validate_columns(lkeys)
        rkeys = lkp_table_obj.validate_columns(rkeys)
        lcolumn_prematch = self.get_columns() # will use this list when matching occurs below
        # create a table to hold the matches
        matches_tobj = self.bing.create_table(into) # create a table to hold the matches
        # for each matching, update it
        for lidx, res in self._scanfuzzy(lkeys, lkp_table_obj, rkeys, min_ratios):
            print('lidx:',lidx,'res:',res)
            for rank, match_row in enumerate(res, start = 1):
                matches_tobj.insert(['lidx', 'ridx', 'ratio', 'rank'], [lidx, match_row[0], match_row[1], rank])
        return        


    def flookup(self, 
                lkp_table: object, 
                on: list[tuple], 
                ret_columns: list[str] | list[tuple],
                ret_matches: bool = False
                ):      
        lkp_table_obj = None
        if isinstance(lkp_table,Memory_Table):
            lkp_table_obj = lkp_table
            lkp_table = lkp_table.name
        else:
            lkp_table_obj = self.bing[lkp_table]
        
        # validate input eg. column etc
        lkeys = [x[0] for x in on] # generate lkeys from on (1st sequence)
        rkeys = [x[1] for x in on] # generate rkeys from on (2nd sequence)
        min_ratios = [x[2] for x in on] # generate ratios from on (3rd sequence)
        lkeys = self.validate_columns(lkeys)
        rkeys = lkp_table_obj.validate_columns(rkeys)
        lcolumn_prematch = self.get_columns() # will use this list when matching occurs below
        # validate ret_columns once:
        # otherwise will make lots of call later
        valid_ret_columns = []
        for item in ret_columns:
            if isinstance(item,str):
                item = lkp_table_obj.validate_column(item)
                valid_ret_columns.append(item)
            if isinstance(item, tuple):
                valid_column  = lkp_table_obj.validate_column(item[0])
                item_ = tuple([valid_column,item[1]])
                valid_ret_columns.append(item_)
        # add these valid_ret_column to the left table
        for col in valid_ret_columns:
            if isinstance(col, str):
                self.add_column(col)
            if isinstance(col, tuple):
                self.add_column(col[1]) # use the alias
        
        # create a table to hold the matches
        # matches_tobj = self.bing.create_table(into) # create a table to hold the matches
        # for each matching, update it
        for lidx, res in self._scanfuzzy(lkeys, lkp_table_obj, rkeys, min_ratios):
            print('lidx:',lidx,'res:',res)
            # if ret_matches:
            #     for rank, match_row in enumerate(res, start = 1):
            #         matches_tobj.insert(['lidx', 'ridx', 'ratio', 'rank'], [lidx, match_row[0], match_row[1], rank])
            # update this table lrow for each ret_columns
            for item in valid_ret_columns:
                if isinstance(item, str): # if item is a column
                    value = lkp_table_obj[res[0][0]][item]
                    if item in lcolumn_prematch:
                        #print('item',item,'in',self.name)
                        self.update_row(lidx, lkp_table_obj.name + '_' + item, value)
                    else:
                        self.update_row(lidx, item, value)
                if isinstance(item, tuple): # if a tuple (0=column to return from lkp_table, 1=as_column)
                    value = lkp_table_obj[res[0][0]][item[0]]
                    self.update_row(lidx, item[1], value)                        
    

    def groupby(self, 
                    columns, 
                    drop_none: bool = True, 
                    group_count: bool = False, 
                    counts: list[str] | list[tuple] = None,
                    sums: list[str] | list[tuple] = None,
                    mins: list[str] | list[tuple] = None,
                    maxs: list[str] | list[tuple] = None,
                    means: list[str] | list[tuple] = None,
                    group_concat = None) -> None:
        """ Generate a groupby records.
            return Table object."""
        group_tobj = Memory_Table('grouped') # table to hold the result
        group_count_column = 'group_count' if group_count == True else group_count
        # loop the table
        valid_cols = [self.validate_column(col) for col in columns] # validate these columns
        for idx, row in self.iterrows():
            index_records = [row[col] for col in valid_cols]
            if drop_none:
                if index_records.count(None) == len(index_records):
                    continue
            index = tuple([bintang.core._normalize_caseless(col) if isinstance(col, str) else col for col in index_records])
            if not group_tobj.index_exists(index):
                # add record for the first time
                group_tobj.insert(valid_cols, index_records, index=index)
                if group_count:
                    group_tobj.update_row(index, group_count_column, 1)
                if group_concat:
                    if group_concat == True:
                        group_tobj.update_row(index, 'group_concat',[idx])  
                    else:
                        group_tobj.update_row(index, 'group_concat',[row[self.validate_column(group_concat)]])        
                if counts:
                    self._groupby_new_index_count(index, row, counts, group_tobj) 
                if sums:
                    self._groupby_new_index_sum(index, row, sums, group_tobj)
                if mins:
                    self._groupby_new_index_min(index, row, mins, group_tobj)
                if maxs:
                    self._groupby_new_index_max(index, row, maxs, group_tobj)
                if means: # call existing sum and count function
                    temp_sum_columns = self._groupby_new_index_sum(index, row, means, group_tobj, means=True)
                    temp_count_columns = self._groupby_new_index_count(index, row, means, group_tobj, means=True)    
            else:
                if group_count:
                    incremented = group_tobj[index][group_count_column] + 1
                    group_tobj.update_row(index, group_count_column, incremented)
                if group_concat:
                    if group_concat == True:
                        group_tobj.update_row(index, 'group_concat', group_tobj[index]['group_concat'] + [idx])
                    else:
                        group_tobj.update_row(index, 'group_concat', group_tobj[index]['group_concat'] + [row[self.validate_column(group_concat)]])        
                if counts:
                    self._groupby_existing_index_count(index, row, counts, group_tobj)
                if sums:
                    self._groupby_existing_index_sum(index, row, sums, group_tobj)
                if mins:
                    self._groupby_existing_index_min(index, row, mins, group_tobj)
                if mins:
                    self._groupby_existing_index_max(index, row, maxs, group_tobj)
                if means: # call sum and count functions
                    self._groupby_existing_index_sum(index, row, means, group_tobj, means=True)
                    self._groupby_existing_index_count(index, row, means, group_tobj, means=True)
        # loop through the groupped table)
        if means:
            for idx, row in group_tobj.iterrows():
                for i, col in enumerate(means):
                    if isinstance(col, str):
                        col_mean = 'mean_' + col
                        group_tobj.update_row(idx, col_mean, row[temp_sum_columns[i]]/row[temp_count_columns[i]])
                    if isinstance(col, tuple):
                        col_mean = col[1]
                        group_tobj.update_row(idx, col_mean, row[temp_sum_columns[i]]/row[temp_count_columns[i]])
            # drop temp columns
            for x in range (len(temp_sum_columns)):
                group_tobj.drop_column(temp_sum_columns[x])
                group_tobj.drop_column(temp_count_columns[x])
        return group_tobj            



  
    def _groupby_new_index_count(self, index, row, counts, group_tobj, means=False):
        count_columns = []
        for col in counts:
            if isinstance(col, str):
                col_cnt = 'count_' + col
                if row[col] is not None:
                    group_tobj.update_row(index, col_cnt, 1)
                else:
                    group_tobj.update_row(index, col_cnt, 0)
                count_columns.append(col_cnt)    
            if isinstance(col, tuple):
                col_cnt = 'mean_count_' + col[1] if means else col[1]
                if row[col[0]] is not None:
                    group_tobj.update_row(index, col_cnt, 1)
                else:
                    group_tobj.update_row(index, col_cnt, 0)
                count_columns.append(col_cnt)
        return count_columns        


    def _groupby_existing_index_count(self, index, row, counts, group_tobj, means=False):
        for col in counts:
            if isinstance(col, str):
                col_cnt = 'count_' + col
                if row[col] is not None:
                    incremented = group_tobj[index][col_cnt] + 1
                    group_tobj.update_row(index, col_cnt, incremented)  
            if isinstance(col, tuple):
                col_cnt = 'mean_count_' + col[1] if means else col[1]
                if row[col[0]] is not None:
                    incremented = group_tobj[index][col_cnt] + 1
                    group_tobj.update_row(index, col_cnt, incremented)

                

    def _groupby_new_index_sum(self, index, row, sums, group_tobj, means=False):
        sum_columns = []
        for col in sums:
            if isinstance(col, str):
                col_sum = 'sum_' + col
                if isinstance(row[col], (int,float)):
                    group_tobj.update_row(index, col_sum, row[col])
                else:
                    group_tobj.update_row(index, col_sum, 0)
                sum_columns.append(col_sum)    
            if isinstance(col, tuple):
                col_sum = 'mean_sum_' + col[1] if means else col[1]
                if isinstance(row[col[0]], (int,float)):
                    group_tobj.update_row(index, col_sum, row[col[0]])
                else:
                    group_tobj.update_row(index, col_sum, 0)
                sum_columns.append(col_sum)
        return sum_columns


    def _groupby_existing_index_sum(self, index, row, sums, group_tobj, means=False):
        for col in sums:
            if isinstance(col, str):
                col_sum = 'sum_' + col
                if isinstance(row[col], (int,float)):
                    summed = group_tobj[index][col_sum] + row[col]
                    group_tobj.update_row(index, col_sum, summed)
            if isinstance(col, tuple):
                col_sum = 'mean_sum_' + col[1] if means else col[1]
                if isinstance(row[col[0]], (int,float)):
                    summed = group_tobj[index][col_sum] + row[col[0]]
                    group_tobj.update_row(index, col_sum, summed)


    def _groupby_new_index_min(self, index, row, mins, group_tobj):
        for col in mins:
            if isinstance(col, str):
                col_min = 'min_' + col
                if isinstance(row[col], (int,float)):
                    group_tobj.update_row(index, col_min, row[col])
                else:
                    group_tobj.update_row(index, col_min, 0)
            if isinstance(col, tuple):
                col_min = col[1]
                if isinstance(row[col[0]], (int,float)):
                    group_tobj.update_row(index, col_min, row[col[0]])
                else:
                    group_tobj.update_row(index, col_min, 0)


    def _groupby_existing_index_min(self, index, row, mins, group_tobj):
        for col in mins:
            if isinstance(col, str):
                col_min = 'min_' + col
                if isinstance(row[col], (int,float)):
                    minned = min(group_tobj[index][col_min], row[col])
                    group_tobj.update_row(index, col_min, minned)
            if isinstance(col, tuple):
                col_sum = col[1]
                if isinstance(row[col[0]], (int,float)):
                    minned = min(group_tobj[index][col_min], row[col[0]])
                    group_tobj.update_row(index, col_min, minned)


    def _groupby_new_index_max(self, index, row, maxs, group_tobj):
        for col in maxs:
            if isinstance(col, str):
                col_max = 'max_' + col
                if isinstance(row[col], (int,float)):
                    group_tobj.update_row(index, col_max, row[col])
                else:
                    group_tobj.update_row(index, col_max, 0)
            if isinstance(col, tuple):
                col_max = col[1]
                if isinstance(row[col[0]], (int,float)):
                    group_tobj.update_row(index, col_max, row[col[0]])
                else:
                    group_tobj.update_row(index, col_max, 0)


    def _groupby_existing_index_max(self, index, row, maxs, group_tobj):
        for col in maxs:
            if isinstance(col, str):
                col_max = 'max_' + col
                if isinstance(row[col], (int,float)):
                    maxed = max(group_tobj[index][col_max], row[col])
                    group_tobj.update_row(index, col_max, maxed)
            if isinstance(col, tuple):
                col_max = col[1]
                if isinstance(row[col[0]], (int,float)):
                    maxed = max(group_tobj[index][col_max], row[col[0]])
                    group_tobj.update_row(index, col_max, maxed)


    def read_csv(self, path, delimiter=',', quotechar='"', header_row=1):
        import csv
        with open(path,newline='') as f:
            reader = csv.reader(f, delimiter=delimiter, quotechar=quotechar)
            # determine columns
            columns = []
            for rownum, row in enumerate(reader, start=1):
                if rownum == header_row:
                    columns = [col for col in row] # add all columns
                    f.seek(0) # return to BOF
                    break 
            
            # insert records
            for rownum, row in enumerate(reader, start=1):
                if rownum > header_row: # assume records after header
                    if len(columns) == len(row):
                        self.insert(columns, row)
                    else:
                        raise IndexError ('length of column and row is not the same at rownum {}. Possible issues were incorrect quoting or missing value.'.format(rownum))


    def iterrows_ws(self, ws, wb_type, header_row=1):
        if wb_type == 'openpyxl':
            for rownum, row_cells in enumerate(ws.iter_rows(min_row=header_row), start=1):
                yield rownum, row_cells
        else: # assume 'xlrd'
            rownum = 1
            for i, row_cells in enumerate(ws, start=1):
                if i < header_row: # simulate 'skiprow' like openpyxl
                    continue # skip
                else:
                    yield rownum, row_cells
                    rownum += 1


    def read_excel(self, wb, sheetname, header_row=1):
        wb_type = bintang.get_wb_type_toread(wb)
        
        # validate sheetname
        if wb_type == 'openpyxl':
            sheetnames_lced = {x.lower(): x  for x in wb.sheetnames}
        else: # assume xlrd
            sheetnames_lced = {x.lower(): x  for x in wb.sheet_names()}

        if sheetname.lower() not in sheetnames_lced:
            similar_sheetnames = bintang.get_similar_values(sheetname, sheetnames_lced)
            raise ValueError ('could not find sheetname {}. Did you mean {}?'.format(repr(sheetname),' or '.join(similar_sheetnames)))
        ws = wb[sheetnames_lced[sheetname.lower()]] # assign with the correct name (caseless) through validated user input.
        columns = []
        Nonecolumn_cnt = 0
        #for rownum, row_cells in enumerate(ws.iter_rows(min_row=header_row),start=1):
        for rownum, row_cells in self.iterrows_ws(ws, wb_type, header_row):
            values = [] # hold column value for each row
            if rownum == 1:
                for cell in row_cells:
                    if cell.value is None:
                        columname = 'noname' + str(Nonecolumn_cnt)
                        Nonecolumn_cnt += 1
                        columns.append(columname)
                    else:
                        columns.append(str(cell.value))
                if Nonecolumn_cnt > 0:
                    log.warning('Warning! Noname column detected!')          
            if rownum > 1:
                for cell in row_cells:
                    values.append(cell.value)
                # if rownum == 370:
                #     log.debug(f'{values} at rownum 370.')
                #     log.debug(any(values))
                if any(values):
                    self.insert(columns, values)


    def read_sql(self, conn, sql_str=None, params=None ):
        cursor = conn.cursor()
        if sql_str is None:
            sql_str = "SELECT * FROM {}".format(self.name)
        if params is not None:
            cursor.execute(sql_str, params)
        else:
            cursor.execute(sql_str)
        columns = [col[0] for col in cursor.description]
        for col in columns:
            self.add_column(col)
        # for row in cursor.fetchall():
        #     self.insert(columns, row) 
        while True:
            rows = cursor.fetchmany(300)
            if not rows: break
            for row in rows:
                self.insert(columns, row)                  


    def reindex(self):
        # dict is already indexed with python 3.7+
        # this only means we rebuild row id so any deleted idx can be reused
        idxs = list(self.__rows)
        for i, idx in enumerate(idxs, 1):
            self.__rows[i] = self.__rows[idx] # reassign
            if i != idx:
                del self.__rows[idx]


    def set_index(self, column):
        grouped = self.groupby([column], group_count=True, group_concat=True)
        temp_test_dup = grouped.filter(lambda row: row['group_count']>1)
        if len(temp_test_dup) > 0:
            # temp_test_dup.name = 'duplicates'
            # temp_test_dup.print()
            dups = Memory_Table('duplicates')
            for idx, row in temp_test_dup.iterrows():
                for _idx in row['group_concat']:
                    dups.insert(self.get_row_asdict(_idx))
            dups.print()
            raise ValueError(f'Duplicates found in column "{column}".')
            #raise ValueError('column must be a single column string!')  
            for idx, row in temp_test_dup.iterrows():
                print(idx, row)
        
        # reset idx using existing column
        temp_tobj = copy.deepcopy(self)
        temp_idxs = list(temp_tobj.__rows)
        # reset table methods
        self.__rows.clear()     # reset all rows
        self.__columns.clear()  # reset all columns
        self.__last_assigned_columnid= 0 #
        self.__last_assigned_rowid = 0 # for use when row created
        self.__last_assigned_idx = 0 # for use when add
        # populate records
        for i in temp_idxs:
            self.insert(temp_tobj.get_row_asdict(i), index=temp_tobj.get_row_asdict(i)[column])
            temp_tobj.delete_row(i)
                    


    def to_csv(self, path, columns=None, index=False, \
               dialect='excel', delimiter=',', \
               quotechar='"', quoting=0): 
        import csv
        # csv.QUOTE_MINIMAL = 0
        # csv.QUOTE_ALL = 1
        # csv.QUOTE_NONNUMERIC = 2
        # csv.QUOTE_NONE = 3
        
        if columns is None:
            columns = self.get_columns()
        with open(path, 'w', newline = '\n') as csvfile:
            csvwriter = csv.writer(csvfile, dialect=dialect, delimiter=delimiter,
                                   quotechar=quotechar, quoting=quoting)
            columns_towrite = [col for col in columns]
            if index:                       # if column index wanted
                idx_col = self.INDEX_COLUMN_NAME
                if isinstance(index, str):  # if user wanted own index column name
                    idx_col = index
                columns_towrite.insert(0,idx_col)
            # write header to csvfile
            csvwriter.writerow(columns_towrite)
            if index:
                for idx, row in self.iterrows(columns, row_type='list'):
                    csvwriter.writerow([idx] + row)
            else:
                for idx, row in self.iterrows(columns, row_type='list'):
                    csvwriter.writerow(row)
            

    def to_dict(self, columns=None):
        if columns is None:
            columns = self.get_columns()
        res = {}
        res['table'] = self.name
        res['columns'] = columns
        res['rows'] = []
        # add row
        for idx, row in self.iterrows(columns):
            res['rows'].append(row)
        return res    
            

    def to_excel_moved_tobase(self, wb, path, columns=None, index=False, sheet_title=None):
        wb_type = bintang.get_wb_type_towrite(wb)
        sheet_title = self.name if sheet_title is None else sheet_title
        if wb_type == 'openpyxl':
            ws = wb.active
            ws.title = sheet_title
        else: # assume xlwt as user want to save as xls
            ws = wb.add_sheet(sheet_title)

        # add header
        columns_towrite = []
        if columns is None:
            columns_towrite = self.get_columns()
        else:
            columns_towrite = [col for col in columns]
        # log.debug('index: {}'.format(index))
        if index:
            if index:                          # if column index wanted
                idx_column = self.INDEX_COLUMN_NAME
                if isinstance(index, str):     # if user wanted own index column name
                    idx_column = index      
                columns_towrite.insert(0,idx_column)
            if index != True:
                columns_towrite.insert(0,str(index))        
        # add headers' row
        if wb_type == 'openpyxl':
            ws.append(columns_towrite)
        else: # assume xlwt
            for i, col in enumerate(columns_towrite):
                ws.write(0, i, col)
        # add data row
        for idx, row in self.iterrows(columns, row_type='list'):
            if index:
                row.insert(0,idx)
            if wb_type == 'openpyxl':
                ws.append(row)
            else: # assume xlwt
                for cidx, value in enumerate(row): # enumerate column value
                    ws.write(idx, cidx, value)     # write value for each column

        # if index:
        #     for idx, row in self.iterrows(columns, row_type='list'):
        #         row.insert(0,idx)
        #         ws.append(row)
        # if not index:
        #     for idx, row in self.iterrows(columns, row_type='list'):
        #         ws.append(row)          
        wb.save(path)


    def to_list(self, column):
        if isinstance(column, str):
            res_list = []
            for idx, row in self.iterrows([column]):
                res_list.append(next(iter(row.values())))
            return res_list    
        else:
            raise ValueError('column must be a single column string!')    


    def _to_json_file(self, path, columns=None):
        """ this will be useful for big table
        and we are going to write row by row into a file
        instead a one-oof write"""
        if columns is None:
            columns = self.get_columns()
        with open(path, 'w') as f:
            f.write('{') # write opening json obj
            f.write('{}'.format(json.dumps(self.name)))
            f.write(':')
            f.write('[') # write opening array

            for idx, row in self.iterrows(columns=columns):
                print(json.dumps(row))
                f.write(json.dumps(row))
                if not (idx + 1) == len(self): 
                    f.write(',') # if not eof then write obj seperator
            f.write(']') # write closing array
            f.write('}') # write closing json obj


    def to_json(self):
        """This is just a placeholder.
        Converting dict to Json is trivia by using Json dumps/dump.
        No need a function to write here?"""
        rows_ = []
        for idx, row in self.iterrows():
            rows_.append(row)
        obj_dict  = {}
        obj_dict[self.name] = rows_
        return json.dumps(obj_dict)



class Path_Table(Memory_Table):
    def __init__(self, name, bing=None):
        super().__init__(name, bing=bing)
        self.path = name
        self.children = [] 
               

    def __repr__(self):
        tbl = {}
        tbl['name'] = self.name
        tbl['path'] = self.path
        columns = []
        for k,v in self._Table__columns.items():
            columns.append(dict(id=v.id, name=v.name))
        tbl['columns'] = columns
        return json.dumps(tbl, indent=2)


    def get_path_aslist(self):
        path_as_list = []
        if self.path == '/': # if root
            path_as_list.append('/')
            return path_as_list
        for node in self.path.split("/"):
            if node == '': # if the 1st one. 1st node have discarded by split
                path_as_list.append('/')
            else:
                path_as_list.append(node)
        return path_as_list


    def upsert_table_path_row(self, tprow):
        #log.debug("\n  ------------------in upsert_table_path_row (table.py) --------------------")
        # extract the rowid and use it as the table index (the key of rows{})
        # create a row if the rowid not found in the table's index
        res_idx = self.get_rowidx_byrowid(tprow.id)
        
        if res_idx is None:
            #log.debug(f'inserting... row does not exist {tprow}')
            row = self.make_row(tprow.id)
            # re-make cells from tprow
            for id, c in tprow.cells.items():
                #log.debug(f'{id} cell: {c}')
                if c.is_key == True:
                    cell = self.make_cell(self.PARENT_PREFIX + c.get_column(), c.value)
                else:
                    cell = self.make_cell(c.get_column(), c.value)

                row.add_cell(cell)
            # add to rows
            self.add_row(row)
            
        elif res_idx is not None:
            #log.debug(f"updating... row exists {tprow}")
            for id, c in tprow.cells.items():
                if c.is_key == True:
                    cell = self.make_cell(self.PARENT_PREFIX + c.get_column(), c.value)
                else:
                    cell = self.make_cell(c.get_column(), c.value)
                # self.__rows[res_idx].add_cell(cell)  
                self._Memory_Table__rows[res_idx].add_cell(cell)        
        #log.debug("\n  ------------------out upsert_table_path_row (table.py)-------------------")        



import json
# import sqlite3
import uuid
import re
import sys
import copy
from pathlib import Path
from typing import Callable
import warnings
import inspect
from itertools import product
from abc import ABC, abstractmethod
from bintang.log import log
import bintang
import types
from bintang.row import Row

MAX_ROW_SQL_INSERT = 300

def custom_formatwarning(msg, *args, **kwargs):
    # ignore everything except the message
    return str(msg)

warnings.formatwarning = custom_formatwarning


class ColumnError(Exception):
    def __init__(self,message):
        self.message = message
    


class Base_Table(ABC):
    """
    Base class for all table types.
    This class provides the basic structure and methods for table manipulation. 
    It is not intended to be instantiated directly, but rather to be subclassed.
    """
    def __init__(self,name, bing=None):
        self.bing: bintang.Bintang = bing # reference to parent Bintang object
        self.name: str = name
        self.INDEX_COLUMN_NAME: str = 'idx'
        self.PARENT_PREFIX: str = ''
        self.__last_assigned_rowid = 0
        self.type_map: dict = type_map # assign type_map to self.type_map
        #self.pytype: dict = pytype # assign pytype to self.pytype

    
    @abstractmethod
    def get_columns(self):
        pass

    @abstractmethod
    def get_columnid(self):
        pass

    @abstractmethod
    def get_columnids(self):
        pass

    @abstractmethod
    def iterrows(self):
        pass

    def ColumnError(self):
        """Raise ColumnError with table name and column name."""
        return ColumnError(self.name, self.column)
    

    def make_row(self,id=None, option=None):
        """make a new row.
        by default it increments id.
        """
        if id is None:
            row = Row(self.__last_assigned_rowid + 1)
            self.__last_assigned_rowid += 1 #increment rowid
        elif id is not None:
            row = Row(id)
        # elif id is None and option == 'uuid':
        #     row = Row(uuid.uuid4())
        return row

    
    def _get_columnnames_lced(self, columns=None) -> dict:
        return {col.lower(): col  for col in self.get_columns()}


    def validate_columns(self, columns):
        """return columns from those stored in table.columns"""
        validated_cols = []
        unmatched_cols = []
        for column in columns:
            if column.lower() in self._get_columnnames_lced().keys():
                validated_cols.append(self._get_columnnames_lced().get(column.lower()))
            else:
                unmatched_cols.append(column)        
        if len(unmatched_cols) > 0:
            similar_columns = self._suggest_similar_columns(unmatched_cols)
            message = self._suggest_columns_msg(similar_columns)
            raise ColumnError(message)
        else:
            return validated_cols 


    def _suggest_similar_columns(self, columns, min_ratio=0.7):
        res = {}
        for col in columns:
            similar_cols = bintang.get_similar_values(col, self.get_columns())
            res[col] = ['{}'.format(x) for x in similar_cols]
        return res  


    def _suggest_columns_msg(self, suggested_columns):   
        unmatched_cols = [x for x in suggested_columns.keys()]
        if len(suggested_columns) > 0:
            message = f"No such {', '.join(unmatched_cols)} column{'(s)' if len(suggested_columns)>1 else ''} in {self.name} table."
            line_msg = []
            for col, suggestion  in suggested_columns.items():
                msg = f" For {col}, did you mean: {', '.join(suggestion)}?"
                line_msg.append(msg)
            # construct message
            for msg in line_msg:
                message += msg
            return message                 

    
    def get_schema_name(self) -> str | None:
        """ Returns the schema of the table.
        Schema is the second part from the right of the table name.
        For example, if the table name is 'public.Person', it returns 'public'.
        """
        parts = self.name.split('.')
        if len(parts) > 1: 
            return parts[-2]


    def extract_schema_name(self, fqtn: str) -> str | None:
        """
        Extracts the schema from a fully qualified table name (fqtn).
        Schema name is the second part from the right of the fqtn.
        For example, if fqtn is 'public.Person', it returns 'public'.
        """
        parts = fqtn.split('.')
        if len(parts) > 1: 
            return parts[-2]
             

    def get_table_name(self) -> str | None:
        parts = self.name.split('.')
        return parts[-1]


    def extract_table_name(self, fqtn) -> str | None:
        parts = fqtn.split('.')
        return parts[-1]


    def _get_sql_conn_name_xp(self, conn):
        # experimenting to get connector name
        # so code can act differently for eg. params list used in prepared statement.
        if str(type(conn)) == "<class 'sqlite3.Connection'>":
            return 'sqlite3'
        elif str(type(conn)) == "<class 'pyodbc.Connection'>":
            return 'pyodbc'
        elif str(type(conn)) == "<class 'psycopg.Connection'>":
            return 'psycopg'
        elif str(type(conn)) == "<class 'psycopg2.extensions.connection'>":
            return 'psycopg2'
        else:
            raise ValueError('Sorry Only sqlite3, pyodbc and psycopg connection accepted!')

            
    def get_value(self, column, where = None):
        """ 
        return a scalar value.
        switch different process if user want column pass as string or lambda
        params:
        column: either column name or using lambda expression
        where: always a lambda expression"""
        if where is not None:
            if isinstance(column,str):
                for idx, row in self.iterrows():
                    if where(row):
                        return row[column]
            elif isinstance(column, types.FunctionType):
                for idx, row in self.iterrows():
                    if where(row):
                        return column(row)
            else:
                raise 'param column must be either column name or lambda and param where must be a lambda.'
        else:
            if isinstance(column,str):
                for idx, row in self.iterrows():
                    return row[column]
            elif isinstance(column, types.FunctionType):
                for idx, row in self.iterrows():
                    return column(row)
            else:
                raise 'param column must be either column name or lambda and param where must be a lambda.'           


    def cmprows(self, lkp_table: str, on: list[tuple] = None, min_matches=1, find_all=True):
        print('len of on', on)
        """
        compare rows in the current table with rows in the lkp_table.
        Args:
            lkp_table (str): the name of the table to compare with aka right table, the index called 'ridx'.
            on (list[str]): a list of tuples where each tuple contains two column names to compare.
                            If None, it will compare all columns in both tables.
            min_matches (int): number of matches to complete when on is None. Default is 1
            find_all (bool): if True (default) then check all the lookup data, while  If False then exit on first match
        Yields:
            lidx, [(ridx,(lcol1,rcol1), ((lcol1,rcol1))]
            1 [(2, ((11, 11), (12, 12)))]
        """
        if on:
            req_matches = len(on)
        else:
            lcolumns = self.get_columns()
            rcolumns = self.bing[lkp_table].get_columns()
            on = list(product(lcolumns,rcolumns))
            req_matches = min_matches 
        on_ = [] # will hold column id keys instead column name
        for tup in on:
            tup_ = (self.get_columnid(tup[0]), self.bing[lkp_table].get_columnid(tup[1]))
            on_.append(tup_) 
        for lidx, lrow in self._iterrows():
            results = []
            for ridx, rrow in self.bing[lkp_table]._iterrows():
                matches = 0
                matched_columns = []
                col_results = []
                for i in range(len(on_)):
                    if on_[i][0] in lrow.cells and on_[i][1] in rrow.cells: # check if both cells exist
                        if bintang.match(lrow.cells[on_[i][0]].value
                                        ,rrow.cells[on_[i][1]].value
                                        ):
                            matches += 1
                            matched_columns.append((on_[i][0] 
                                                ,on_[i][1]
                                                ))
                            cmp = (on_[i][0], on_[i][1])                    
                            col_results.append(cmp)
                if matches >= req_matches:
                    results.append((ridx,tuple(col_results)))
                    #yield lidx, ridx, tuple(matched_columns)
                    if not find_all:
                        break  # only return the first match
            if results: # yield only when the results is True/not empty
                yield lidx, results    
    
    
    def to_excel(self, wb, path, columns=None, index=False, sheet_title=None):
        wb_type = bintang.get_wb_type_towrite(wb)
        sheet_title = self.name if sheet_title is None else sheet_title
        if wb_type == 'openpyxl':
            ws = wb.active
            ws.title = sheet_title
        else: # assume xlwt as user want to save as xls
            ws = wb.add_sheet(sheet_title)

        # add header
        columns_towrite = []
        if columns is None:
            columns_towrite = self.get_columns()
        else:
            columns_towrite = [col for col in columns]
        # log.debug('index: {}'.format(index))
        if index:
            if index:                          # if column index wanted
                idx_column = self.INDEX_COLUMN_NAME
                if isinstance(index, str):     # if user wanted own index column name
                    idx_column = index      
                columns_towrite.insert(0,idx_column)
            if index != True:
                columns_towrite.insert(0,str(index))        
        # add headers' row
        if wb_type == 'openpyxl':
            ws.append(columns_towrite)
        else: # assume xlwt
            for i, col in enumerate(columns_towrite):
                ws.write(0, i, col)
        # add data row
        for idx, row in self.iterrows(columns, row_type='list'):
            if index:
                row.insert(0,idx)
            if wb_type == 'openpyxl':
                ws.append(row)
            else: # assume xlwt
                for cidx, value in enumerate(row): # enumerate column value
                    ws.write(idx, cidx, value)     # write value for each column

        # if index:
        #     for idx, row in self.iterrows(columns, row_type='list'):
        #         row.insert(0,idx)
        #         ws.append(row)
        # if not index:
        #     for idx, row in self.iterrows(columns, row_type='list'):
        #         ws.append(row)          
        wb.save(path)
    
    
    def to_sql(self, conn: object, 
               table: str, 
               columns: list[str]=None,
               method: str='prepared', 
               max_rows: int = 1) -> int:
        """conn: accept only pyodbc.Connection or psycopg.Connection
           table: table name in the database and can use a fully qualified table name eg. schema_name.table_name.
           columns: If a dictionary then a columns mapping where the key is sql column (destination) and the value is bintang columns (source). 
                    If a list, column mapping will be created automatically assuming source columns and destination columns are the same. 
                    If not provided it assumes that user wants to insert all the columns from the table.
           method: prepared (default) or string
           return-> rows_affected: number of rows affected by the insert operation.
        """
        conn_name = self._get_sql_conn_name_xp(conn)
        if columns is None: # check if user want to include all columns in the table
            columns = self.get_columns()
        else:
            if isinstance(columns,list) or isinstance(columns, tuple):
                columns= self.validate_columns(columns)
            elif isinstance(columns,dict):
                columns_key = [x for x in columns.keys()]
                columns_val = [x for x in columns.values()] ## trouble maker
                columns_val= self.validate_columns(columns_val) ## trouble maker
                columns = dict(zip(columns_key, columns_val))
            else: 
                raise ValueError('Error! Only list or dict allowed for columns.')

        schema = self.extract_schema_name(table)
        table = self.extract_table_name(table)

        if method.upper() =='STRING':
            return self._to_sql_string(conn, table, columns, schema=schema, max_rows=max_rows, conn_name=conn_name)
        else:
            return self._to_sql_prep(conn, table, columns, schema=schema, max_rows=max_rows,conn_name=conn_name)

 
    def _to_sql_string(self, conn, table, columns, schema=None, max_rows = 300, conn_name='pyodbc'):
        colmap = self.set_to_sql_colmap(columns)
        src_cols = [x for x in colmap.values()]
        dest_columns = [x for x in colmap.keys()]
        if schema is not None:
            sql_template = 'INSERT INTO "{}"."{}" ({}) VALUES'
            str_stmt = sql_template.format(schema,table,",".join(['"{}"'.format(x) for x in colmap]))
        else:
            sql_template = 'INSERT INTO "{}" ({}) VALUES'
            str_stmt = sql_template.format(table,",".join(['"{}"'.format(x) for x in colmap]))
        
        # check if getting type info is supported, if yes then get sql data type and literal
        res = None
        try:
            res = self.get_sql_typeinfo_table(conn)
        except Exception as e:
            log.error(e)
            log.warning('Getting SQL type info is not supported for this connection, will proceed without getting SQL data type and literal. This may cause issue for some data types and may cause all values to be treated as string literal.')
        if res is not None:
            sql_cols_withtype = self.set_sql_datatype(dest_columns, conn, schema, table)
            sql_cols_withliteral = self.set_sql_literal(sql_cols_withtype, conn)
        else:
            sql_cols_withliteral = None

        cursor = conn.cursor()
        temp_rows = []  
        rows_affected = 0 # to hold total record affected
        for idx, values in self.iterrows(src_cols, row_type='list'):
            if sql_cols_withliteral is not None:    
                sql_record = self.gen_sql_record_with_literal(values, sql_cols_withliteral)
            else:   
                sql_record = self.gen_sql_record_without_literal(values)
            temp_rows.append(sql_record)
            if len(temp_rows) == max_rows:
                stmt = str_stmt + ' {}'.format(",".join(temp_rows))
                try:
                    log.debug(stmt)
                    cursor.execute(stmt)
                    rows_affected += cursor.rowcount
                except Exception as e:
                    log.error(e)
                    log.error(stmt)
                temp_rows.clear()
        if len(temp_rows) > 0:
            stmt = str_stmt + ' {}'.format(",".join(temp_rows))
            try:
                log.debug(stmt)
                cursor.execute(stmt)
                rows_affected += cursor.rowcount
            except Exception as e:
                log.error(e)
                log.error(stmt)
        return rows_affected           


    def gen_sql_record_with_literal(self, values, sql_cols_withliteral):
        sql_record = []
        for (col, value) in zip(sql_cols_withliteral.keys(), values):
            literals = sql_cols_withliteral[col]
            sql_value = self.gen_sql_value_with_literal(value, literals)
            sql_record.append(sql_value)
        return "({})".format(','.join(sql_record))


    def gen_sql_record_without_literal(self, values):
        sql_record = []
        for value in values:
            sql_value = self.gen_sql_value_without_literal(value)
            sql_record.append(sql_value)
        return "({})".format(','.join(sql_record))    
        

    def gen_sql_value_with_literal(self, value, literals):
        if value == "" or value is None:
            return "NULL"
        if isinstance(value, str):
            value = value.replace("'","''")
        return f"{'' if literals[0] is None else literals[0]}{value}{'' if literals[1] is None else literals[1]}"


    def gen_sql_value_without_literal(self, value):
        if value == "" or value is None:
            return "NULL"
        if isinstance(value, str):
            value = value.replace("'","''")
        
        if type(value) in [int, float, bool]:
            return str(value) # so we can just join them later
        else:
            return f"'{value}'"


    def get_sql_typeinfo_table(self, conn):
        cursor = conn.cursor()
        sql_type_info_tuple = cursor.getTypeInfo(sqlType = None)
        columns_ = [column[0] for column in cursor.description]
        sql_typeinfo = {}
        for row in sql_type_info_tuple:
            row_dict = dict(zip(columns_, row))
            type_name = row_dict['type_name']     
            sql_typeinfo[type_name] = row_dict
        return sql_typeinfo
    

    def set_sql_datatype(self, dest_columns, conn, schema, table):
        cursor = conn.cursor()
        try:
            sql_columns = cursor.columns(schema=schema, table=table)
            columns_ = [column[0] for column in cursor.description]
            col_dict = {}
            for row in sql_columns: 
                row_dict = dict(zip(columns_, row))
                column_name = row_dict['column_name']
                col_dict[row_dict['column_name']] = row_dict
            sql_columns_withtype = {}    
            for col in dest_columns:
                sql_columns_withtype[col] = col_dict[col]['type_name']    
            return sql_columns_withtype
        except Exception as e: 
            log.error(e)
            return None  
    

    def set_sql_literal(self, sql_cols_withtype, conn):
        try:
            sql_typeinfo_tab = self.get_sql_typeinfo_table(conn)           
            sql_cols_withliteral = {}
            for k, v in sql_cols_withtype.items():
                prefix = sql_typeinfo_tab[v]['literal_prefix']
                suffix = sql_typeinfo_tab[v]['literal_suffix']
                literals = (prefix,suffix)
                sql_cols_withliteral[k] = literals
            return sql_cols_withliteral    
        except Exception as e: # opps... no support
            log.error(e)
            literals = (None, None)
            return {col:literals for col in sql_cols_withtype.keys()}
        


    def _to_sql_prep(self, conn, table, columns, schema=None, max_rows = 1, conn_name='pyodbc'):
        if max_rows <= len(self): # validate max_row
            mrpb = max_rows # assign max row per batch
        else:
            #log.warning('Warning! max_rows {} set greater than totalrows {}. max_rows set to 1'.format(max_rows, len(self)))
            mrpb = 1
        numof_col = len(columns) # num of columns
        colmap = self.set_to_sql_colmap(columns)
        src_cols = [x for x in colmap.values()]
        dest_columns = [x for x in colmap.keys()]
        
        # generate prepared statement
        prep_stmt = self.gen_prep_stmt(table, dest_columns, numof_col, mrpb, conn_name=conn_name, schema=schema)
        
        cursor = conn.cursor()
        temp_rows = []
        rows_affected = 0
        for idx, row in self.iterrows(src_cols, row_type='list'):
            for v in row:
                temp_rows.append(v)
            if len(temp_rows) == (mrpb * numof_col):
                log.debug(prep_stmt)
                log.debug(temp_rows)
                try:
                    cursor.execute(prep_stmt, temp_rows)
                    rows_affected += cursor.rowcount
                except Exception as e:
                    log.error(e)
                    log.error(prep_stmt)
                    log.error(temp_rows)  
                temp_rows.clear()     
        
        if len(temp_rows) > 0: # if any reminder
            # create as prepared statement
            mrpb = int(len(temp_rows)/numof_col) # adjust mrpb
            prep_stmt = self.gen_prep_stmt(table, dest_columns, numof_col, mrpb, conn_name=conn_name, schema=schema)
            try:
                cursor.execute(prep_stmt, temp_rows)
                rows_affected += cursor.rowcount
            except Exception as e:
                log.error(e)
                log.error(prep_stmt)
                log.error(temp_rows)
        return rows_affected


    def gen_prep_stmt(self, table, dest_columns, numof_col, mrpb, conn_name='pyodbc', schema=None):
        param_markers = self.gen_row_param_markers(numof_col, mrpb, conn_name=conn_name)
        if schema is not None:
            sql_template = 'INSERT INTO "{}"."{}" ({}) VALUES {}'
            prep_stmt = sql_template.format(schema, table, ",".join(['"{}"'.format(x) for x in dest_columns]),param_markers)
        else:
            sql_template = 'INSERT INTO "{}" ({}) VALUES {}'
            prep_stmt = sql_template.format(table, ",".join(['"{}"'.format(x) for x in dest_columns]),param_markers)
        return prep_stmt
                

    def gen_row_param_markers(self,numof_col,num_row, conn_name='pyodbc'):
        p = "%s" if conn_name[:7] =='psycopg' else "?" # param marker is different for pyodbc and psycopg, so we need to check the connection name
        param = "(" + ",".join([p]  *numof_col) + ")"
        params = []
        for i in range(num_row):
            params.append(param)
        return ",".join(params)

        
    def update(self, column, value, where=None):
        if isinstance(value, types.FunctionType):
            # value passed as function
            for idx, row in self.iterrows():
                if where is None:
                    try:
                        val = value(row) # function called
                        self.update_row(idx, column, val)
                    except Exception as e:
                        #log.warning(e)
                        pass    
                else:    
                    try:
                        if where(row):
                            val = value(row) # function called
                            self.update_row(idx, column, val) 
                    except Exception as e:
                        #log.warning(e)
                        pass 
        else:
            # value passed as non function eg. an int or str
            for idx, row in self.iterrows():
                if where is None:
                    self.update_row(idx, column, value)
                else:    
                    try:
                        if where(row): # function called 
                            self.update_row(idx, column, value) 
                    except Exception as e:
                        # print(e)
                        pass 


    def gen_sql_stmt_merge_dev(self
                           ,trg_table
                           ,dbms = None
                           ,key_columns = None
                           ,proc_name = None
                           ):
        SQU = type_map[dbms]['delimited_identifiers']['start']
        EQU = type_map[dbms]['delimited_identifiers']['end']
        columns = self.get_columns()

        # generate 'on' clause
        on_colmap_str = '<TO BE DEFINED>'
        on_colmap = []
        if key_columns is not None:
            for col in key_columns:
                item = self.quote('src.' + col, dbms) + ' = ' + self.quote('trg.' + col, dbms)       
                on_colmap.append(item)
            on_colmap_str = ' AND '.join(on_colmap)

        # generate colmap for update clause
        update_colmap = []
        for col in columns:
            if col not in key_columns:
                item = self.quote(col, dbms) + ' = ' + self.quote('src', dbms) + '.' + self.quote(col, dbms) + '\n\t\t'
                update_colmap.append(item)
        update_colmap_str = ','.join(update_colmap)

        merge_stmt = f"""
        MERGE INTO {self.quote(trg_table, dbms)} AS trg
        USING {self.quote(self.name, dbms)} AS src
        ON ({on_colmap_str})
        WHEN MATCHED THEN
            UPDATE SET {update_colmap_str}
        WHEN NOT MATCHED BY TARGET THEN
            INSERT ({','.join([self.quote(col, dbms) for col in columns])})
            VALUES ({','.join([self.quote('src', dbms) + '.' + self.quote(col, dbms) for col in columns])})
        WHEN NOT MATCHED BY SOURCE THEN
            DELETE
        ;
        """
        # gen proc_name
        schema_str = ''
        schema = self.get_schema_name()
        if schema is not None:
            schema_str = self.quote(schema, dbms) + '.'

        proc_name_str = self.quote(f'Merge_{self.get_table_name()}_To_TRG', dbms)
        proc_name = f'{schema_str}{proc_name_str}' if proc_name is None else 'proc_name'

        proc_stmt = f"""
        CREATE OR ALTER PROCEDURE {proc_name}
        AS
        BEGIN
        {merge_stmt}
        END
        """
        return proc_stmt


    def quote(self, name, dbms):
        STARTQ = type_map[dbms]['delimited_identifiers']['start']
        ENDQ = type_map[dbms]['delimited_identifiers']['end']
        split_name = name.split('.')
        return '.'.join([STARTQ + x + ENDQ for x in split_name])


    def gen_sql_stmt_truncate_table_dev(self, dbms, proc_name = None):
        schema_str = ''
        schema = self.get_schema_name()
        if schema is not None:
            schema_str = self.quote(schema, dbms) + '.'
        
        truncate_stmt = f"""
        TRUNCATE TABLE {self.quote(self.name, dbms)}
        """
        proc_name_str = self.quote(f'Truncate_Table_{self.get_table_name()}', dbms)
        proc_name = f'{schema_str}{proc_name_str}' if proc_name is not None else 'proc_name'
        proc_stmt = f"""
        CREATE OR ALTER PROCEDURE {proc_name}
        AS
        BEGIN
        {truncate_stmt}
        END
        """
        return proc_stmt         



type_map = {
    'sqlite': {
        'type_mappings': {
            'str':'text'
            ,'int':'int'
            ,'datetime':'datetime'
            ,'float':'real'
            ,'bool':'int'
            ,'bytes':'blob'
        },
        'delimited_identifiers':{'start':'"', 'end':'"'},
        'type_info': {
        'varchar': {'literal_prefix':"'", 'literal_suffix':"'"}
        } 
    },
    'sqlserver': {
        'type_mappings': {
            'str':'nvarchar'
            ,'int':'int'
            ,'datetime':'datetime'
            ,'float':'float'
            ,'bool':'bit'
            ,'bytes':'varbinary'
        },
        'delimited_identifiers':{'start':'[', 'end':']'},
        'type_info': {
        'varchar': {'literal_prefix':"'", 'literal_suffix':"'"}
        } 
    },
    'postgresql': {
        'type_mappings': {
            'str':'varchar'
            ,'int':'INTEGER'
            ,'datetime':'timestamp'
            ,'float':'float8'
            ,'bool':'boolean'
            ,'bytes':'bytes'
        },
        'delimited_identifiers':{'start':'"', 'end':'"'},
        'type_info': {
            'varchar': {'literal_prefix':"'", 'literal_suffix':"'"}
        }   
    }            
}




